import os
import base64
import magic
from typing import Dict


def check_is_web_url(file_path: str):
    for url_scheme in ["https://", "http://"]:
        if file_path.startswith(url_scheme):
            return True
    return False


def get_attachment_json_for_file(file_path: str, file_name: str, ignore_if_error: bool) -> Dict:
    # Ensure that path is expanded and absolute
    abs_path = os.path.abspath(os.path.expanduser(file_path))
    # Get attachment json
    try:
        with open(abs_path, "rb") as f:
            final_file_name = os.path.basename(abs_path)
            if file_name and file_name.strip():
                final_file_name = file_name.strip()
            # --
            mime_type = magic.from_file(abs_path, mime=True)
            # base64 encoded string
            b64encoded = base64.b64encode(f.read())
            b64data = b64encoded.decode()
            attach_data = {
                "filename": final_file_name,
                "contentType": mime_type,
                "data": b64data,
                "url": None,
                "ignore_if_error": ignore_if_error,
            }
            return attach_data
    except OSError as ex:
        if ignore_if_error:
            print("WARNING: ignoring error while processing attachment file.", ex)
            return None
        else:
            raise ex


def get_attachment_json_for_url(file_url: str, file_name: str, ignore_if_error: bool) -> Dict:
    return {
        "filename": file_name,
        "contentType": None,
        "data": None,
        "url": file_url,
        "ignore_if_error": ignore_if_error,
    }


def get_attachment_json(file_path: str, file_name: str = None, ignore_if_error: bool = False) -> Dict:
    if check_is_web_url(file_path):
        return get_attachment_json_for_url(file_path, file_name, ignore_if_error)
    else:
        return get_attachment_json_for_file(file_path, file_name, ignore_if_error)



import hashlib
import hmac
import base64
import json
from typing import Dict, Tuple
from urllib.parse import urlparse


def get_request_signature(url: str, http_verb: str, content, headers: Dict, secret: str) -> Tuple[str, str]:
    if http_verb == "GET":  # POST/GET/PUT
        content_txt, content_md5 = "", ""
    else:
        if content == "":
            content_txt, content_md5 = "", ""
        else:
            content_txt = json.dumps(content, ensure_ascii=False)
            content_md5 = hashlib.md5(content_txt.encode()).hexdigest()
    # ----
    request_uri = get_uri(url)
    # ----- Create string to sign
    string_to_sign = "{}\n{}\n{}\n{}\n{}".format(
        http_verb,
        content_md5,
        headers["Content-Type"],
        headers["Date"],
        request_uri
    )
    # print("string_to_sign", string_to_sign)
    # ----- HMAC-SHA-256
    sig_hexdigest = hmac.HMAC(secret.encode(), msg=string_to_sign.encode(), digestmod=hashlib.sha256).digest()
    # -----
    sig = base64.b64encode(sig_hexdigest).decode()  # decode('utf-8'/'ascii')
    return content_txt, sig


def get_uri(url: str) -> str:
    o_url = urlparse(url)
    request_uri = o_url.path
    if o_url.query:
        request_uri = "{}?{}".format(request_uri, o_url.query)

    return request_uri



import logging
from http.client import HTTPConnection
# Enabling debugging at http.client level (requests->urllib3->http.client)
# you will see the REQUEST, including HEADERS and DATA, and RESPONSE with HEADERS but without DATA.
# the only thing missing will be the response.body which is not logged.
requests_log = None

logging.basicConfig()  # you need to initialize logging, otherwise you will not see anything from requests


def set_logging(level=0):
    global requests_log
    if level > 0:
        HTTPConnection.debuglevel = 1
        #
        logging.getLogger().setLevel(logging.DEBUG)
        requests_log = logging.getLogger("urllib3")
        requests_log.setLevel(logging.DEBUG)
        requests_log.propagate = True
    else:
        HTTPConnection.debuglevel = 0
        #
        logging.getLogger().setLevel(logging.WARN)
        requests_log = logging.getLogger("urllib3")
        requests_log.setLevel(logging.WARN)
        requests_log.propagate = True



from datetime import datetime, timezone
import requests
from typing import Dict
import urllib.parse

from .exception import SuprsendAPIException, SuprsendValidationError
from .constants import HEADER_DATE_FMT
from .signature import get_request_signature


class TenantsApi:
    def __init__(self, config):
        self.config = config
        self.list_url = self.__list_url()
        self.__headers = self.__common_headers()

    def __list_url(self):
        list_uri_template = "{}v1/tenant/"
        list_uri_template = list_uri_template.format(self.config.base_url)
        return list_uri_template

    def __common_headers(self):
        return {
            "Content-Type": "application/json; charset=utf-8",
            "User-Agent": self.config.user_agent,
        }

    def __dynamic_headers(self):
        return {
            "Date": datetime.now(timezone.utc).strftime(HEADER_DATE_FMT),
        }

    def cleaned_limit_offset(self, limit: int, offset: int):
        # limit must be 0 < x <= 1000
        limit = limit if (isinstance(limit, int) and 0 < limit <= 1000) else 20
        # offset must be >=0
        offset = offset if (isinstance(offset, int) and offset >= 0) else 0
        #
        return limit, offset

    def list(self, limit: int = 20, offset: int = 0):
        limit, offset = self.cleaned_limit_offset(limit, offset)
        params = {"limit": limit, "offset": offset}
        encoded_params = urllib.parse.urlencode(params)
        #
        url = f"{self.list_url}?{encoded_params}"
        # ---
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'GET', None, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.get(url, headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def _validate_tenant_id(self, tenant_id):
        if not isinstance(tenant_id, (str,)):
            raise SuprsendValidationError("tenant_id must be a string")
        tenant_id = tenant_id.strip()
        if not tenant_id:
            raise SuprsendValidationError("missing tenant_id")
        return tenant_id

    def detail_url(self, tenant_id: str):
        tenant_id_encoded = urllib.parse.quote_plus(tenant_id)
        url = f"{self.list_url}{tenant_id_encoded}/"
        return url

    def get(self, tenant_id: str):
        tenant_id = self._validate_tenant_id(tenant_id)
        url = self.detail_url(tenant_id)
        # ---
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'GET', None, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.get(url, headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def upsert(self, tenant_id: str, tenant_payload: Dict):
        tenant_id = self._validate_tenant_id(tenant_id)
        url = self.detail_url(tenant_id)
        # ---
        tenant_payload = tenant_payload or {}
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'POST', tenant_payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.post(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def delete(self, tenant_id: str):
        tenant_id = self._validate_tenant_id(tenant_id)
        url = self.detail_url(tenant_id)
        # ---
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'DELETE', "", headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.delete(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return {"success": True, "status_code": resp.status_code}



from typing import Any, Dict, Iterable, Union
import time
import uuid

from .constants import (
    IDENTITY_SINGLE_EVENT_MAX_APPARENT_SIZE_IN_BYTES,
    IDENTITY_SINGLE_EVENT_MAX_APPARENT_SIZE_IN_BYTES_READABLE,
)
from .exception import InputValueError
from .utils import (get_apparent_identity_event_size, )
from .user_edit_internal_helper import _UserEditInternalHelper


class UserEdit:
    def __init__(self, config, distinct_id: str):
        self.config = config
        self.distinct_id = distinct_id
        #
        self.__errors = []
        self.__info = []
        #
        self.operations = []
        self._helper = _UserEditInternalHelper()
        #
        self.__warnings_list = []

    @property
    def warnings(self):
        return self.__info

    @property
    def errors(self):
        return self.__errors

    def get_payload(self):
        return {"operations": self.operations}

    def get_async_payload(self):
        return {
            "$schema": "2",
            "$insert_id": str(uuid.uuid4()),
            "$time": int(time.time() * 1000),
            "env": self.config.workspace_key,
            "distinct_id": self.distinct_id,
            "$user_operations": self.operations,
            "properties": {"$ss_sdk_version": self.config.user_agent},
        }

    def as_json_async(self):
        return {
            "distinct_id": self.distinct_id,
            "$user_operations": self.operations,
            "warnings": self.__warnings_list,
        }

    def validate_payload_size(self, payload: Dict):
        apparent_size = get_apparent_identity_event_size(payload)
        if apparent_size > IDENTITY_SINGLE_EVENT_MAX_APPARENT_SIZE_IN_BYTES:
            raise InputValueError(f"User Payload size too big - {apparent_size} Bytes, "
                                  f"must not cross {IDENTITY_SINGLE_EVENT_MAX_APPARENT_SIZE_IN_BYTES_READABLE}")
        # ----
        return payload, apparent_size

    def validate_body(self):
        self.__warnings_list = []
        if self.__info:
            msg = f"[distinct_id: {self.distinct_id}]" + "\n".join(self.__info)
            self.__warnings_list.append(msg)
            # print on console as well
            print(f"WARNING: {msg}")
        if self.__errors:
            msg = f"[distinct_id: {self.distinct_id}]" + "\n".join(self.__errors)
            self.__warnings_list.append(msg)
            print(f"ERROR: {msg}")
        # ------
        return self.__warnings_list

    def _collect_operation(self):
        resp = self._helper.get_operation_result()
        if resp["errors"]:
            self.__errors.extend(resp["errors"])
        if resp["info"]:
            self.__info.extend(resp["info"])
        if resp["operation"]:
            self.operations.append(resp["operation"])

    def append(self, arg1: Union[str, Dict], arg2: Any = None):
        """
        Usage:
        1. append(k, v)
        2. append({k1: v1, k2, v2})

        :param arg1: one of [str, dict]
        :param arg2: required if arg1 is string
        :return:
        """
        caller = "append"
        if not isinstance(arg1, (str, dict)):
            self.__errors.append(f"[{caller}] arg1 must be either string or a dict")
            return
        if isinstance(arg1, (str,)):
            if arg2 is None:
                self.__errors.append(f"[{caller}] if arg1 is a string, then arg2 must be passed")
                return
            else:
                self._helper._append_kv(arg1, arg2, {}, caller=caller)
                self._collect_operation()
        else:
            for k, v in arg1.items():
                self._helper._append_kv(k, v, arg1, caller=caller)
            # --
            self._collect_operation()

    def set(self, arg1: Union[str, Dict], arg2: Any = None):
        """
        1. set(k, v)
        2. set({k1: v1, k2, v2})

        :param arg1: required, one of  [str, dict]
        :param arg2: required if arg1 is str
        :return:
        """
        caller = "set"
        if not isinstance(arg1, (str, dict)):
            self.__errors.append(f"[{caller}] arg1 must be String or a dict")
            return
        if isinstance(arg1, (str,)):
            if arg2 is None:
                self.__errors.append(f"[{caller}] if arg1 is a string, then arg2 must be passed")
                return
            else:
                self._helper._set_kv(arg1, arg2, caller=caller)
                self._collect_operation()
        else:
            for k, v in arg1.items():
                self._helper._set_kv(k, v, caller=caller)
            # --
            self._collect_operation()

    def set_once(self, arg1: Union[str, Dict], arg2: Any = None):
        """
        1. set_once(k, v)
        2. set_once({k1: v1, k2, v2})

        :param arg1: required, one of  [str, dict]
        :param arg2: required if arg1 is str
        :return:
        """
        caller = "set_once"
        if not isinstance(arg1, (str, dict)):
            self.__errors.append(f"[{caller}] arg1 must be String or a dict")
            return
        if isinstance(arg1, (str,)):
            if arg2 is None:
                self.__errors.append(f"[{caller}] if arg1 is a string, then arg2 must be passed")
                return
            else:
                self._helper._set_once_kv(arg1, arg2, caller=caller)
                self._collect_operation()
        else:
            for k, v in arg1.items():
                self._helper._set_once_kv(k, v, caller=caller)
            # --
            self._collect_operation()

    def increment(self, arg1: Union[str, Dict], arg2: Any = None):
        """
        1. increment(k, v)
        2. increment({k1: v1, k2, v2})

        :param arg1: required, one of  [str, dict]
        :param arg2: required if arg1 is str
        :return:
        """
        caller = "increment"
        if not isinstance(arg1, (str, dict)):
            self.__errors.append(f"[{caller}] arg1 must be String or a dict")
            return
        if isinstance(arg1, (str,)):
            if arg2 is None:
                self.__errors.append(f"[{caller}] if arg1 is a string, then arg2 must be passed")
                return
            else:
                self._helper._increment_kv(arg1, arg2, caller=caller)
                self._collect_operation()
        else:
            for k, v in arg1.items():
                self._helper._increment_kv(k, v, caller=caller)
            # --
            self._collect_operation()

    def remove(self, arg1: Union[str, Dict], arg2: Any = None):
        """
        Usage:
        1. remove(k, v)
        2. remove({k1: v1, k2, v2})

        :param arg1: one of [str, dict]
        :param arg2: required if arg1 is string
        :return:
        """
        caller = "remove"
        if not isinstance(arg1, (str, dict)):
            self.__errors.append(f"[{caller}] arg1 must be either string or a dict")
            return
        if isinstance(arg1, (str,)):
            if arg2 is None:
                self.__errors.append(f"[{caller}] if arg1 is a string, then arg2 must be passed")
                return
            else:
                self._helper._remove_kv(arg1, arg2, {}, caller=caller)
                self._collect_operation()
        else:
            for k, v in arg1.items():
                self._helper._remove_kv(k, v, arg1, caller=caller)
            # --
            self._collect_operation()

    def unset(self, key: Union[str, Iterable[str]]):
        """
        Usage:
        1. unset(k)
        2. unset([k1, k2])

        :param key: one of [str, list[str]]
        :return:
        """
        caller = "unset"
        if not isinstance(key, (str, list, tuple, set)):
            self.__errors.append(f"[{caller}] key must be either String or List[string]")
            return
        if isinstance(key, (str,)):
            self._helper._unset_k(key, caller=caller)
            self._collect_operation()
        else:
            for k in key:
                self._helper._unset_k(k, caller=caller)
            # --
            self._collect_operation()

    # ------------------------ Preferred language
    def set_preferred_language(self, lang_code: str):
        """
        :param lang_code:
        :return:
        """
        caller = "set_preferred_language"
        self._helper._set_preferred_language(lang_code, caller=caller)
        self._collect_operation()

    # ------------------------ Locale
    def set_locale(self, locale_code: str):
        """
        :param locale_code:
        :return:
        """
        caller = "set_locale"
        self._helper._set_locale(locale_code, caller=caller)
        self._collect_operation()

    # ------------------------ Timezone
    def set_timezone(self, timezone: str):
        """
        :param timezone:
        :return:
        """
        caller = "set_timezone"
        self._helper._set_timezone(timezone, caller=caller)
        self._collect_operation()

    # ------------------------ Email
    def add_email(self, value: str):
        """

        :param value:
        :return:
        """
        caller = "add_email"
        self._helper._add_email(value, caller=caller)
        self._collect_operation()

    def remove_email(self, value: str):
        """

        :param value:
        :return:
        """
        caller = "remove_email"
        self._helper._remove_email(value, caller=caller)
        self._collect_operation()

    # ------------------------ SMS
    def add_sms(self, value: str):
        """

        :param value:
        :return:
        """
        caller = "add_sms"
        self._helper._add_sms(value, caller=caller)
        self._collect_operation()

    def remove_sms(self, value: str):
        """

        :param value:
        :return:
        """
        caller = "remove_sms"
        self._helper._remove_sms(value, caller=caller)
        self._collect_operation()

    # ------------------------ Whatsapp
    def add_whatsapp(self, value: str):
        """

        :param value:
        :return:
        """
        caller = "add_whatsapp"
        self._helper._add_whatsapp(value, caller=caller)
        self._collect_operation()

    def remove_whatsapp(self, value: str):
        """

        :param value:
        :return:
        """
        caller = "remove_whatsapp"
        self._helper._remove_whatsapp(value, caller=caller)
        self._collect_operation()

    # ------------------------ Androidpush
    def add_androidpush(self, value: str, provider: str = None):
        """

        :param value:
        :param provider:
        :return:
        """
        caller = "add_androidpush"
        self._helper._add_androidpush(value, provider, caller=caller)
        self._collect_operation()

    def remove_androidpush(self, value: str, provider: str = None):
        """

        :param value:
        :param provider:
        :return:
        """
        caller = "remove_androidpush"
        self._helper._remove_androidpush(value, provider, caller=caller)
        self._collect_operation()

    # ------------------------ Iospush [providers: apns]
    def add_iospush(self, value: str, provider: str = None):
        """

        :param value:
        :param provider:
        :return:
        """
        caller = "add_iospush"
        self._helper._add_iospush(value, provider, caller=caller)
        self._collect_operation()

    def remove_iospush(self, value: str, provider: str = None):
        """

        :param value:
        :param provider:
        :return:
        """
        caller = "remove_iospush"
        self._helper._remove_iospush(value, provider, caller=caller)
        self._collect_operation()

    # ------------------------ Webpush [providers: vapid]
    def add_webpush(self, value: dict, provider: str = None):
        """

        :param value:
        :param provider:
        :return:
        """
        caller = "add_webpush"
        self._helper._add_webpush(value, provider, caller=caller)
        self._collect_operation()

    def remove_webpush(self, value: dict, provider: str = None):
        """

        :param value:
        :param provider:
        :return:
        """
        caller = "remove_webpush"
        self._helper._remove_webpush(value, provider, caller=caller)
        self._collect_operation()

    # ------------------------ Slack
    def add_slack(self, value: dict):
        """

        :param value:
        :return:
        """
        caller = "add_slack"
        self._helper._add_slack(value, caller=caller)
        self._collect_operation()

    def remove_slack(self, value: dict):
        """

        :param value:
        :return:
        """
        caller = "remove_slack"
        self._helper._remove_slack(value, caller=caller)
        self._collect_operation()

    # ------------------------ MS Teams
    def add_ms_teams(self, value: dict):
        """

        :param value:
        :return:
        """
        caller = "add_ms_teams"
        self._helper._add_ms_teams(value, caller=caller)
        self._collect_operation()

    def remove_ms_teams(self, value: dict):
        """

        :param value:
        :return:
        """
        caller = "remove_ms_teams"
        self._helper._remove_ms_teams(value, caller=caller)
        self._collect_operation()



from datetime import datetime, timezone
from typing import Dict, Union

import requests
import urllib.parse

from .constants import (
    HEADER_DATE_FMT,
)
from .exception import SuprsendAPIException, SuprsendValidationError
from .signature import get_request_signature
from .object_edit import ObjectEdit


class ObjectsApi:
    def __init__(self, config):
        self.config = config
        self.list_url = "{}v1/object/".format(self.config.base_url)
        self.bulk_url = "{}v1/bulk/object/".format(self.config.base_url)

    def __get_headers(self):
        return {
            "Content-Type": "application/json; charset=utf-8",
            "User-Agent": self.config.user_agent,
            "Date": datetime.now(timezone.utc).strftime(HEADER_DATE_FMT),
        }

    def _validate_object_type(self, object_type: str):
        if not object_type or not isinstance(object_type, (str,)) or not object_type.strip():
            raise SuprsendValidationError("missing object_type")
        return object_type.strip()

    def _validate_object_id(self, object_id: str):
        if not object_id or not isinstance(object_id, (str,)) or not object_id.strip():
            raise SuprsendValidationError("missing object_id")
        return object_id.strip()

    def list(self, object_type: str, options: Dict = None) -> Dict:
        object_type = self._validate_object_type(object_type)
        object_type_encoded = urllib.parse.quote_plus(object_type)
        encoded_options = urllib.parse.urlencode((options or {}))
        #
        url = "{}{}/{}".format(self.list_url, object_type_encoded, (f"?{encoded_options}" if encoded_options else ""))
        headers = self.__get_headers()
        # ---
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "GET", None, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.get(url, headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def detail_url(self, object_type: str, object_id: str) -> str:
        object_type = self._validate_object_type(object_type)
        object_type_encoded = urllib.parse.quote_plus(object_type)
        # --
        object_id = self._validate_object_id(object_id)
        object_id_encoded = urllib.parse.quote_plus(object_id)
        # --
        url = f"{self.list_url}{object_type_encoded}/{object_id_encoded}/"
        return url

    def get(self, object_type: str, object_id: str) -> Dict:
        url = self.detail_url(object_type, object_id)
        headers = self.__get_headers()
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "GET", None, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.get(url, headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def upsert(self, object_type: str, object_id: str, payload: Dict = None) -> Dict:
        url = self.detail_url(object_type, object_id)
        payload = payload or {}
        headers = self.__get_headers()
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "POST", payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.post(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def edit(self, edit_ins_or_object_type: Union[ObjectEdit, str], object_id: str = None, edit_payload: Dict = None) -> Dict:
        if isinstance(edit_ins_or_object_type, ObjectEdit):
            edit_ins = edit_ins_or_object_type
            edit_ins.validate_body()
            payload = edit_ins.get_payload()
            url = self.detail_url(edit_ins.object_type, edit_ins.object_id)
        else:
            object_type = edit_ins_or_object_type
            payload = edit_payload or {}
            url = self.detail_url(object_type, object_id)
        # ---
        headers = self.__get_headers()
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "PATCH", payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.patch(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def delete(self, object_type: str, object_id: str) -> Dict:
        url = self.detail_url(object_type, object_id)
        headers = self.__get_headers()
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "DELETE", "", headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.delete(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return {"success": True, "status_code": resp.status_code}

    def bulk_delete(self, object_type: str, payload: Dict) -> Dict:
        """
        payload: {"object_ids": ["id1", "id2"]}
        :param object_type:
        :param payload:
        :return:
        """
        object_type = self._validate_object_type(object_type)
        object_type_encoded = urllib.parse.quote_plus(object_type)
        url = "{}{}/".format(self.bulk_url, object_type_encoded)
        payload = payload or {}
        headers = self.__get_headers()
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "DELETE", payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.delete(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return {"success": True, "status_code": resp.status_code}

    def get_subscriptions(self, object_type: str, object_id: str, options: Dict = None) -> Dict:
        encoded_options = urllib.parse.urlencode((options or {}))
        _detail_url = self.detail_url(object_type, object_id)
        url = "{}subscription/{}".format(_detail_url, (f"?{encoded_options}" if encoded_options else ""))
        headers = self.__get_headers()
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "GET", None, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.get(url, headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def create_subscriptions(self, object_type: str, object_id: str, payload: Dict) -> Dict:
        """
        payload: {
            "recipients": ["distinct_id1", {"object_type": "type1", "id": "id1"},],
            "properties": {"type": "admin"}
        }
        :param object_type:
        :param object_id:
        :param payload:
        :return:
        """
        _detail_url = self.detail_url(object_type, object_id)
        url = "{}subscription/".format(_detail_url)
        payload = payload or {}
        headers = self.__get_headers()
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "POST", payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.post(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def delete_subscriptions(self, object_type: str, object_id: str, payload: Dict) -> Dict:
        """
        payload: {
            "recipients": ["distinct_id1", {"object_type": "type1", "id": "id1"},]
        }
        :param object_type:
        :param object_id:
        :param payload:
        :return:
        """
        _detail_url = self.detail_url(object_type, object_id)
        url = "{}subscription/".format(_detail_url)
        payload = payload or {}
        headers = self.__get_headers()
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "DELETE", payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.delete(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return {"success": True, "status_code": resp.status_code}

    def get_objects_subscribed_to(self, object_type: str, object_id: str, options: Dict = None) -> Dict:
        encoded_options = urllib.parse.urlencode((options or {}))
        _detail_url = self.detail_url(object_type, object_id)
        url = "{}subscribed_to/object/{}".format(_detail_url, (f"?{encoded_options}" if encoded_options else ""))
        headers = self.__get_headers()
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "GET", None, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.get(url, headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def get_edit_instance(self, object_type: str, object_id: str) -> ObjectEdit:
        object_type = self._validate_object_type(object_type)
        object_id = self._validate_object_id(object_id)
        return ObjectEdit(self.config, object_type, object_id)



from typing import Any, Dict, Union

# ---------- Identity keys
IDENT_KEY_EMAIL = "$email"
IDENT_KEY_SMS = "$sms"
IDENT_KEY_ANDROIDPUSH = "$androidpush"
IDENT_KEY_IOSPUSH = "$iospush"
IDENT_KEY_WHATSAPP = "$whatsapp"
IDENT_KEY_WEBPUSH = "$webpush"
IDENT_KEY_SLACK = "$slack"
IDENT_KEY_MS_TEAMS = "$ms_teams"

IDENT_KEYS_ALL = [IDENT_KEY_EMAIL, IDENT_KEY_SMS, IDENT_KEY_ANDROIDPUSH, IDENT_KEY_IOSPUSH,
                  IDENT_KEY_WHATSAPP, IDENT_KEY_WEBPUSH, IDENT_KEY_SLACK, IDENT_KEY_MS_TEAMS]

KEY_ID_PROVIDER = "$id_provider"
KEY_PREFERRED_LANGUAGE = "$preferred_language"
KEY_LOCALE = "$locale"
KEY_TIMEZONE = "$timezone"


class _UserEditInternalHelper:
    def __init__(self):
        self.__dict_set = {}
        self.__dict_set_once = {}
        self.__dict_increment = {}
        self.__dict_append = {}
        self.__dict_remove = {}
        self.__list_unset = []
        #
        self.__errors = []
        self.__info = []

    def reset(self):
        self.__dict_set, self.__dict_append, self.__dict_remove, self.__list_unset = {}, {}, {}, []
        self.__dict_set_once, self.__dict_increment = {}, {}
        self.__errors = []
        self.__info = []

    def get_operation_result(self):
        operation = self.__form_operation()
        ret_val = {
            "errors": self.__errors,
            "info": self.__info,
            "operation": operation,
        }
        self.reset()
        return ret_val

    def __form_operation(self):
        event = {}
        if self.__dict_set:
            event["$set"] = self.__dict_set
        if self.__dict_set_once:
            event["$set_once"] = self.__dict_set_once
        if self.__dict_increment:
            event["$add"] = self.__dict_increment
        if self.__dict_append:
            event["$append"] = self.__dict_append
        if self.__dict_remove:
            event["$remove"] = self.__dict_remove
        if self.__list_unset:
            event["$unset"] = self.__list_unset
        return event

    # ------------------------
    def __validate_key_basic(self, key: str, caller: str):
        if not isinstance(key, (str,)):
            self.__info.append(f"[{caller}] skipping key: {key}. key must be a string")
            return key, False
        key = key.strip()
        if not key:
            self.__info.append(f"[{caller}] skipping key: empty string")
            return key, False
        # -----
        return key, True

    def __is_identity_key(self, key: str):
        return key in IDENT_KEYS_ALL

    # -------------------------
    def _append_kv(self, key: str, val: Any, kwargs: Dict, caller: str = "append"):
        key, is_k_valid = self.__validate_key_basic(key, caller)
        if not is_k_valid:
            return
        if self.__is_identity_key(key):
            self.__add_identity(key, val, kwargs, caller=caller)
        else:
            self.__dict_append[key] = val

    def _remove_kv(self, key: str, val: Any, kwargs: Dict, caller: str = "remove"):
        key, is_k_valid = self.__validate_key_basic(key, caller)
        if not is_k_valid:
            return
        if self.__is_identity_key(key):
            self.__remove_identity(key, val, kwargs, caller=caller)
        else:
            self.__dict_remove[key] = val

    def _unset_k(self, key: str, caller: str = "unset"):
        k, is_k_valid = self.__validate_key_basic(key, caller)
        if not is_k_valid:
            return
        # ----
        self.__list_unset.append(k)

    def _set_kv(self, key: str, val: Any, caller: str = "set"):
        key, is_k_valid = self.__validate_key_basic(key, caller)
        if not is_k_valid:
            return
        else:
            self.__dict_set[key] = val

    def _set_once_kv(self, key: str, val: Any, caller: str = "set_once"):
        key, is_k_valid = self.__validate_key_basic(key, caller)
        if not is_k_valid:
            return
        else:
            self.__dict_set_once[key] = val

    def _increment_kv(self, key: str, val: Union[int, float], caller: str = "increment"):
        key, is_k_valid = self.__validate_key_basic(key, caller)
        if not is_k_valid:
            return
        else:
            self.__dict_increment[key] = val

    def _set_preferred_language(self, lang_code: str, caller: str):
        self.__dict_set[KEY_PREFERRED_LANGUAGE] = lang_code

    def _set_locale(self, locale_code: str, caller: str):
        self.__dict_set[KEY_LOCALE] = locale_code

    def _set_timezone(self, timezone: str, caller: str):
        self.__dict_set[KEY_TIMEZONE] = timezone

    def __add_identity(self, key: str, val: Union[str, Any], kwargs: Dict, caller: str):
        new_caller = f"{caller}:{key}"
        if key == IDENT_KEY_EMAIL:
            self._add_email(val, caller=new_caller)

        elif key == IDENT_KEY_SMS:
            self._add_sms(val, caller=new_caller)

        elif key == IDENT_KEY_WHATSAPP:
            self._add_whatsapp(val, caller=new_caller)

        elif key == IDENT_KEY_ANDROIDPUSH:
            self._add_androidpush(val, kwargs.get(KEY_ID_PROVIDER), caller=new_caller)

        elif key == IDENT_KEY_IOSPUSH:
            self._add_iospush(val, kwargs.get(KEY_ID_PROVIDER), caller=new_caller)

        elif key == IDENT_KEY_WEBPUSH:
            self._add_webpush(val, kwargs.get(KEY_ID_PROVIDER), caller=new_caller)

        elif key == IDENT_KEY_SLACK:
            self._add_slack(val, caller=new_caller)

        elif key == IDENT_KEY_MS_TEAMS:
            self._add_ms_teams(val, caller=new_caller)

    def __remove_identity(self, key: str, val: Union[str, Any], kwargs: Dict, caller: str):
        new_caller = f"{caller}:{key}"
        if key == IDENT_KEY_EMAIL:
            self._remove_email(val, caller=new_caller)

        elif key == IDENT_KEY_SMS:
            self._remove_sms(val, caller=new_caller)

        elif key == IDENT_KEY_WHATSAPP:
            self._remove_whatsapp(val, caller=new_caller)

        elif key == IDENT_KEY_ANDROIDPUSH:
            self._remove_androidpush(val, kwargs.get(KEY_ID_PROVIDER), caller=new_caller)

        elif key == IDENT_KEY_IOSPUSH:
            self._remove_iospush(val, kwargs.get(KEY_ID_PROVIDER), caller=new_caller)

        elif key == IDENT_KEY_WEBPUSH:
            self._remove_webpush(val, kwargs.get(KEY_ID_PROVIDER), caller=new_caller)

        elif key == IDENT_KEY_SLACK:
            self._remove_slack(val, caller=new_caller)

        elif key == IDENT_KEY_MS_TEAMS:
            self._remove_ms_teams(val, caller=new_caller)

    # ------------------------ Email
    def _add_email(self, value: str, caller: str):
        self.__dict_append[IDENT_KEY_EMAIL] = value

    def _remove_email(self, value: str, caller: str):
        self.__dict_remove[IDENT_KEY_EMAIL] = value

    # ------------------------ SMS

    def _add_sms(self, value: str, caller: str):
        self.__dict_append[IDENT_KEY_SMS] = value

    def _remove_sms(self, value: str, caller: str):
        self.__dict_remove[IDENT_KEY_SMS] = value

    # ------------------------ Whatsapp

    def _add_whatsapp(self, value: str, caller: str):
        self.__dict_append[IDENT_KEY_WHATSAPP] = value

    def _remove_whatsapp(self, value: str, caller: str):
        self.__dict_remove[IDENT_KEY_WHATSAPP] = value

    # ------------------------ Androidpush

    def _add_androidpush(self, value: str, provider: str, caller: str):
        self.__dict_append[IDENT_KEY_ANDROIDPUSH] = value
        self.__dict_append[KEY_ID_PROVIDER] = provider

    def _remove_androidpush(self, value: str, provider: str, caller: str):
        self.__dict_remove[IDENT_KEY_ANDROIDPUSH] = value
        self.__dict_remove[KEY_ID_PROVIDER] = provider

    # ------------------------ Iospush

    def _add_iospush(self, value: str, provider: str, caller: str):
        self.__dict_append[IDENT_KEY_IOSPUSH] = value
        self.__dict_append[KEY_ID_PROVIDER] = provider

    def _remove_iospush(self, value: str, provider: str, caller: str):
        self.__dict_remove[IDENT_KEY_IOSPUSH] = value
        self.__dict_remove[KEY_ID_PROVIDER] = provider

    # ------------------------ Webpush

    def _add_webpush(self, value: dict, provider: str, caller: str):
        self.__dict_append[IDENT_KEY_WEBPUSH] = value
        self.__dict_append[KEY_ID_PROVIDER] = provider

    def _remove_webpush(self, value: dict, provider: str, caller: str):
        self.__dict_remove[IDENT_KEY_WEBPUSH] = value
        self.__dict_remove[KEY_ID_PROVIDER] = provider

    # ------------------------ Slack

    def _add_slack(self, value: dict, caller: str):
        self.__dict_append[IDENT_KEY_SLACK] = value

    def _remove_slack(self, value: dict, caller: str):
        self.__dict_remove[IDENT_KEY_SLACK] = value

    # ------------------------ MS Teams

    def _add_ms_teams(self, value: dict, caller: str):
        self.__dict_append[IDENT_KEY_MS_TEAMS] = value

    def _remove_ms_teams(self, value: dict, caller: str):
        self.__dict_remove[IDENT_KEY_MS_TEAMS] = value



"""
Integrate with SuprSend platform using python

For more information about this library, checkout the README on GitHub:
    https://github.com/suprsend/suprsend-py-sdk
For the user guide, examples and more, visit the docs page at:
    https://docs.suprsend.com/docs
"""
from .version import __version__
__author__ = 'SuprSend Developers'
__credits__ = 'SuprSend'

from .sdkinstance import Suprsend           # noqa
from .bulk_response import BulkResponse     # noqa
from .event import Event                    # noqa
from .workflow import Workflow              # noqa
from .workflow_request import WorkflowTriggerRequest # noqa
from .subscriber_list import SubscriberListBroadcast      # noqa
from .object_edit import ObjectEdit
from .user_edit import UserEdit
from .users_edit_bulk import BulkUsersEdit

from .exception import (
    SuprsendError, SuprsendConfigError, SuprsendAPIException, SuprsendValidationError,
    InputValueError,
)



from datetime import datetime, timezone
import requests
import time
from typing import Any, Dict, Iterable, Union
import uuid

from .constants import (
    HEADER_DATE_FMT,
    IDENTITY_SINGLE_EVENT_MAX_APPARENT_SIZE_IN_BYTES,
    IDENTITY_SINGLE_EVENT_MAX_APPARENT_SIZE_IN_BYTES_READABLE,
)
from .exception import InputValueError
from .signature import get_request_signature
from .utils import (get_apparent_identity_event_size, )
from .subscriber_helper import _SubscriberInternalHelper


class SubscriberFactory:
    def __init__(self, config):
        self.config = config

    def new(self, distinct_id: str = None):
        return self.get_instance(distinct_id)

    def get_instance(self, distinct_id: str = None):
        if not isinstance(distinct_id, (str,)):
            raise InputValueError("distinct_id must be a string. an Id which uniquely identify a user in your app")
        distinct_id = distinct_id.strip()
        if not distinct_id:
            raise InputValueError("distinct_id must be passed")
        # -----
        return Subscriber(self.config, distinct_id)


class Subscriber:
    def __init__(self, config, distinct_id: str):
        self.config = config
        self.distinct_id = distinct_id
        self.__url = "{}event/".format(self.config.base_url)
        #
        self.__errors = []
        self.__info = []
        self.user_operations = []
        self._helper = _SubscriberInternalHelper()
        self.__warnings_list = []

    def __get_headers(self):
        return {
            "Content-Type": "application/json; charset=utf-8",
            "Date": datetime.now(timezone.utc).strftime(HEADER_DATE_FMT),
            "User-Agent": self.config.user_agent,
        }

    @property
    def warnings(self):
        return self.__info

    @property
    def errors(self):
        return self.__errors

    def get_event(self):
        return {
            "$schema": "2",
            "$insert_id": str(uuid.uuid4()),
            "$time": int(time.time() * 1000),
            "env": self.config.workspace_key,
            "distinct_id": self.distinct_id,
            "$user_operations": self.user_operations,
            "properties": {"$ss_sdk_version": self.config.user_agent},
        }

    def as_json(self):
        event_dict = {
            "distinct_id": self.distinct_id,
            "$user_operations": self.user_operations,
            "warnings": self.__warnings_list,
        }
        # -----
        return event_dict

    def validate_event_size(self, event_dict: Dict):
        apparent_size = get_apparent_identity_event_size(event_dict)
        if apparent_size > IDENTITY_SINGLE_EVENT_MAX_APPARENT_SIZE_IN_BYTES:
            raise InputValueError(f"User Event size too big - {apparent_size} Bytes, "
                                  f"must not cross {IDENTITY_SINGLE_EVENT_MAX_APPARENT_SIZE_IN_BYTES_READABLE}")
        # ----
        return event_dict, apparent_size

    def validate_body(self, is_part_of_bulk=False):
        self.__warnings_list = []
        if self.__info:
            msg = f"[distinct_id: {self.distinct_id}]" + "\n".join(self.__info)
            self.__warnings_list.append(msg)
            # print on console as well
            print(f"WARNING: {msg}")
        if self.__errors:
            msg = f"[distinct_id: {self.distinct_id}]" + "\n".join(self.__errors)
            self.__warnings_list.append(msg)
            err_msg = f"ERROR: {msg}"
            if is_part_of_bulk:
                # print on console in case of bulk-api
                print(err_msg)
            else:
                # raise error in case of single api
                # raise InputValueError(err_msg)  # Removed exception throwing, let backend handle it
                print(err_msg)
        # ------
        return self.__warnings_list

    def save(self):
        try:
            self.validate_body(is_part_of_bulk=False)
            headers = self.__get_headers()
            event = self.get_event()
            # --- validate event size
            ev, size = self.validate_event_size(event)

            # --- Signature and Authorization-header
            content_txt, sig = get_request_signature(self.__url, 'POST', event, headers,
                                                     self.config.workspace_secret)
            headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
            # -----
            resp = requests.post(self.__url,
                                 data=content_txt.encode('utf-8'),
                                 headers=headers)
        except Exception as ex:
            error_str = ex.__str__()
            return {
                "success": False,
                "status": "fail",
                "status_code": 500,
                "message": error_str,
            }
        else:
            ok_response = resp.status_code // 100 == 2
            if ok_response:
                return {
                    "success": True,
                    "status": "success",
                    "status_code": resp.status_code,
                    "message": resp.text,
                }
            else:
                return {
                    "success": False,
                    "status": "fail",
                    "status_code": resp.status_code,
                    "message": resp.text,
                }

    def _collect_event(self):
        resp = self._helper.get_identity_event()
        if resp["errors"]:
            self.__errors.extend(resp["errors"])
        if resp["info"]:
            self.__info.extend(resp["info"])
        if resp["event"]:
            self.user_operations.append(resp["event"])

    def append(self, arg1: Union[str, Dict], arg2: Any = None):
        """
        Usage:
        1. append(k, v)
        2. append({k1: v1, k2, v2})

        :param arg1: one of [str, dict]
        :param arg2: required if arg1 is string
        :return:
        """
        caller = "append"
        if not isinstance(arg1, (str, dict)):
            self.__errors.append(f"[{caller}] arg1 must be either string or a dict")
            return
        if isinstance(arg1, (str,)):
            if arg2 is None:
                self.__errors.append(f"[{caller}] if arg1 is a string, then arg2 must be passed")
                return
            else:
                self._helper._append_kv(arg1, arg2, {}, caller=caller)
                self._collect_event()
        else:
            for k, v in arg1.items():
                self._helper._append_kv(k, v, arg1, caller=caller)
            # --
            self._collect_event()

    def set(self, arg1: Union[str, Dict], arg2: Any = None):
        """
        1. set(k, v)
        2. set({k1: v1, k2, v2})

        :param arg1: required, one of  [str, dict]
        :param arg2: required if arg1 is str
        :return:
        """
        caller = "set"
        if not isinstance(arg1, (str, dict)):
            self.__errors.append(f"[{caller}] arg1 must be String or a dict")
            return
        if isinstance(arg1, (str,)):
            if arg2 is None:
                self.__errors.append(f"[{caller}] if arg1 is a string, then arg2 must be passed")
                return
            else:
                self._helper._set_kv(arg1, arg2, caller=caller)
                self._collect_event()
        else:
            for k, v in arg1.items():
                self._helper._set_kv(k, v, caller=caller)
            # --
            self._collect_event()

    def set_once(self, arg1: Union[str, Dict], arg2: Any = None):
        """
        1. set_once(k, v)
        2. set_once({k1: v1, k2, v2})

        :param arg1: required, one of  [str, dict]
        :param arg2: required if arg1 is str
        :return:
        """
        caller = "set_once"
        if not isinstance(arg1, (str, dict)):
            self.__errors.append(f"[{caller}] arg1 must be String or a dict")
            return
        if isinstance(arg1, (str,)):
            if arg2 is None:
                self.__errors.append(f"[{caller}] if arg1 is a string, then arg2 must be passed")
                return
            else:
                self._helper._set_once_kv(arg1, arg2, caller=caller)
                self._collect_event()
        else:
            for k, v in arg1.items():
                self._helper._set_once_kv(k, v, caller=caller)
            # --
            self._collect_event()

    def increment(self, arg1: Union[str, Dict], arg2: Any = None):
        """
        1. increment(k, v)
        2. increment({k1: v1, k2, v2})

        :param arg1: required, one of  [str, dict]
        :param arg2: required if arg1 is str
        :return:
        """
        caller = "increment"
        if not isinstance(arg1, (str, dict)):
            self.__errors.append(f"[{caller}] arg1 must be String or a dict")
            return
        if isinstance(arg1, (str,)):
            if arg2 is None:
                self.__errors.append(f"[{caller}] if arg1 is a string, then arg2 must be passed")
                return
            else:
                self._helper._increment_kv(arg1, arg2, caller=caller)
                self._collect_event()
        else:
            for k, v in arg1.items():
                self._helper._increment_kv(k, v, caller=caller)
            # --
            self._collect_event()

    def remove(self, arg1: Union[str, Dict], arg2: Any = None):
        """
        Usage:
        1. remove(k, v)
        2. remove({k1: v1, k2, v2})

        :param arg1: one of [str, dict]
        :param arg2: required if arg1 is string
        :return:
        """
        caller = "remove"
        if not isinstance(arg1, (str, dict)):
            self.__errors.append(f"[{caller}] arg1 must be either string or a dict")
            return
        if isinstance(arg1, (str,)):
            if arg2 is None:
                self.__errors.append(f"[{caller}] if arg1 is a string, then arg2 must be passed")
                return
            else:
                self._helper._remove_kv(arg1, arg2, {}, caller=caller)
                self._collect_event()
        else:
            for k, v in arg1.items():
                self._helper._remove_kv(k, v, arg1, caller=caller)
            # --
            self._collect_event()

    def unset(self, key: Union[str, Iterable[str]]):
        """
        Usage:
        1. unset(k)
        2. unset([k1, k2])

        :param key: one of [str, list[str]]
        :return:
        """
        caller = "unset"
        if not isinstance(key, (str, list, tuple, set)):
            self.__errors.append(f"[{caller}] key must be either String or List[string]")
            return
        if isinstance(key, (str,)):
            self._helper._unset_k(key, caller=caller)
            self._collect_event()
        else:
            for k in key:
                self._helper._unset_k(k, caller=caller)
            # --
            self._collect_event()

    # ------------------------ Preferred language
    def set_preferred_language(self, lang_code: str):
        """
        :param lang_code:
        :return:
        """
        caller = "set_preferred_language"
        self._helper._set_preferred_language(lang_code, caller=caller)
        self._collect_event()

    # ------------------------ Locale
    def set_locale(self, locale_code: str):
        """
        :param locale_code:
        :return:
        """
        caller = "set_locale"
        self._helper._set_locale(locale_code, caller=caller)
        self._collect_event()

    # ------------------------ Timezone
    def set_timezone(self, timezone: str):
        """
        :param timezone:
        :return:
        """
        caller = "set_timezone"
        self._helper._set_timezone(timezone, caller=caller)
        self._collect_event()

    # ------------------------ Email
    def add_email(self, value: str):
        """

        :param value:
        :return:
        """
        caller = "add_email"
        self._helper._add_email(value, caller=caller)
        self._collect_event()

    def remove_email(self, value: str):
        """

        :param value:
        :return:
        """
        caller = "remove_email"
        self._helper._remove_email(value, caller=caller)
        self._collect_event()

    # ------------------------ SMS
    def add_sms(self, value: str):
        """

        :param value:
        :return:
        """
        caller = "add_sms"
        self._helper._add_sms(value, caller=caller)
        self._collect_event()

    def remove_sms(self, value: str):
        """

        :param value:
        :return:
        """
        caller = "remove_sms"
        self._helper._remove_sms(value, caller=caller)
        self._collect_event()

    # ------------------------ Whatsapp
    def add_whatsapp(self, value: str):
        """

        :param value:
        :return:
        """
        caller = "add_whatsapp"
        self._helper._add_whatsapp(value, caller=caller)
        self._collect_event()

    def remove_whatsapp(self, value: str):
        """

        :param value:
        :return:
        """
        caller = "remove_whatsapp"
        self._helper._remove_whatsapp(value, caller=caller)
        self._collect_event()

    # ------------------------ Androidpush
    def add_androidpush(self, value: str, provider: str = None):
        """

        :param value:
        :param provider:
        :return:
        """
        caller = "add_androidpush"
        self._helper._add_androidpush(value, provider, caller=caller)
        self._collect_event()

    def remove_androidpush(self, value: str, provider: str = None):
        """

        :param value:
        :param provider:
        :return:
        """
        caller = "remove_androidpush"
        self._helper._remove_androidpush(value, provider, caller=caller)
        self._collect_event()

    # ------------------------ Iospush [providers: apns]
    def add_iospush(self, value: str, provider: str = None):
        """

        :param value:
        :param provider:
        :return:
        """
        caller = "add_iospush"
        self._helper._add_iospush(value, provider, caller=caller)
        self._collect_event()

    def remove_iospush(self, value: str, provider: str = None):
        """

        :param value:
        :param provider:
        :return:
        """
        caller = "remove_iospush"
        self._helper._remove_iospush(value, provider, caller=caller)
        self._collect_event()

    # ------------------------ Webpush [providers: vapid]
    def add_webpush(self, value: dict, provider: str = None):
        """

        :param value:
        :param provider:
        :return:
        """
        caller = "add_webpush"
        self._helper._add_webpush(value, provider, caller=caller)
        self._collect_event()

    def remove_webpush(self, value: dict, provider: str = None):
        """

        :param value:
        :param provider:
        :return:
        """
        caller = "remove_webpush"
        self._helper._remove_webpush(value, provider, caller=caller)
        self._collect_event()

    # ------------------------ Slack
    def add_slack(self, value: dict):
        """

        :param value:
        :return:
        """
        caller = "add_slack"
        self._helper._add_slack(value, caller=caller)
        self._collect_event()

    def remove_slack(self, value: dict):
        """

        :param value:
        :return:
        """
        caller = "remove_slack"
        self._helper._remove_slack(value, caller=caller)
        self._collect_event()

    # ------------------------ MS Teams
    def add_ms_teams(self, value: dict):
        """

        :param value:
        :return:
        """
        caller = "add_ms_teams"
        self._helper._add_ms_teams(value, caller=caller)
        self._collect_event()

    def remove_ms_teams(self, value: dict):
        """

        :param value:
        :return:
        """
        caller = "remove_ms_teams"
        self._helper._remove_ms_teams(value, caller=caller)
        self._collect_event()



from datetime import datetime, timezone
import requests
from typing import List, Dict
import urllib.parse

from .exception import SuprsendAPIException
from .constants import HEADER_DATE_FMT
from .signature import get_request_signature


class BrandsApi:
    def __init__(self, config):
        self.config = config
        self.list_url = self.__list_url()
        self.__headers = self.__common_headers()

    def __list_url(self):
        list_uri_template = "{}v1/brand/"
        list_uri_template = list_uri_template.format(self.config.base_url)
        return list_uri_template

    def __common_headers(self):
        return {
            "Content-Type": "application/json; charset=utf-8",
            "User-Agent": self.config.user_agent,
        }

    def __dynamic_headers(self):
        return {
            "Date": datetime.now(timezone.utc).strftime(HEADER_DATE_FMT),
        }

    def cleaned_limit_offset(self, limit: int, offset: int):
        # limit must be 0 < x <= 1000
        limit = limit if (isinstance(limit, int) and 0 < limit <= 1000) else 20
        # offset must be >=0
        offset = offset if (isinstance(offset, int) and offset >= 0) else 0
        #
        return limit, offset

    def list(self, limit: int = 20, offset: int = 0):
        limit, offset = self.cleaned_limit_offset(limit, offset)
        params = {"limit": limit, "offset": offset}
        encoded_params = urllib.parse.urlencode(params)
        #
        url = f"{self.list_url}?{encoded_params}"
        # ---
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'GET', None, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.get(url, headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def detail_url(self, brand_id: str):
        brand_id = str(brand_id).strip()
        brand_id_encoded = urllib.parse.quote_plus(brand_id)
        url = f"{self.list_url}{brand_id_encoded}/"
        return url

    def get(self, brand_id: str):
        url = self.detail_url(brand_id)
        # ---
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'GET', None, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.get(url, headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def upsert(self, brand_id: str, brand_payload: Dict):
        url = self.detail_url(brand_id)
        # ---
        brand_payload = brand_payload or {}
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'POST', brand_payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.post(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()



from datetime import datetime, timezone
import requests
from typing import Dict
from warnings import warn

from .constants import (
    HEADER_DATE_FMT,
    BODY_MAX_APPARENT_SIZE_IN_BYTES, BODY_MAX_APPARENT_SIZE_IN_BYTES_READABLE,
)
from .exception import InputValueError
from .utils import (get_apparent_workflow_body_size, validate_workflow_body_schema)
from .signature import get_request_signature
from .attachment import get_attachment_json


class Workflow:
    def __init__(self, body, idempotency_key: str = None, tenant_id: str = None, brand_id: str = None):
        # warn(f'{self.__class__.__name__} will be deprecated.', DeprecationWarning, stacklevel=2)
        if not isinstance(body, (dict,)):
            raise InputValueError("workflow body must be a json/dictionary")
        self.body = body
        self.idempotency_key = idempotency_key
        self.tenant_id = tenant_id
        self.brand_id = brand_id

    def add_attachment(self, file_path: str, file_name: str = None, ignore_if_error: bool = False):
        if self.body.get("data") is None:
            self.body["data"] = {}
        # if body["data"] is not a dict, not raising error while adding attachment.
        if not isinstance(self.body["data"], (dict,)):
            print("WARNING: attachment cannot be added. please make sure body['data'] is a dictionary. Workflow" +
                  str(self.as_json()))
            return
        # ---
        attachment = get_attachment_json(file_path, file_name, ignore_if_error)
        if not attachment:
            return
        # --- add the attachment to body->data->$attachments
        if self.body["data"].get("$attachments") is None:
            self.body["data"]["$attachments"] = []
        # -----
        self.body["data"]["$attachments"].append(attachment)

    def get_final_json(self, config, is_part_of_bulk: bool = False):
        # add idempotency key in body if present
        if self.idempotency_key:
            self.body["$idempotency_key"] = self.idempotency_key
        if self.tenant_id:
            self.body["tenant_id"] = self.tenant_id
        if self.brand_id:
            self.body["brand_id"] = self.brand_id
        # --
        self.body = validate_workflow_body_schema(self.body)
        # ---- Check body size
        apparent_size = get_apparent_workflow_body_size(self.body, is_part_of_bulk)
        if apparent_size > BODY_MAX_APPARENT_SIZE_IN_BYTES:
            raise InputValueError(f"workflow body too big - {apparent_size} Bytes, "
                                  f"must not cross {BODY_MAX_APPARENT_SIZE_IN_BYTES_READABLE}")
        # ----
        return self.body, apparent_size

    def as_json(self):
        body_dict = {**self.body}
        if self.idempotency_key:
            body_dict["$idempotency_key"] = self.idempotency_key
        if self.tenant_id:
            body_dict["tenant_id"] = self.tenant_id
        if self.brand_id:
            body_dict["brand_id"] = self.brand_id
        # -----
        return body_dict


class _WorkflowTrigger:
    def __init__(self, config):
        self.config = config
        self.url = self.__get_url()

    def __get_url(self):
        url_formatted = "{}{}/trigger/".format(self.config.base_url, self.config.workspace_key)
        return url_formatted

    def __get_headers(self):
        return {
            "Content-Type": "application/json; charset=utf-8",
            "Date": datetime.now(timezone.utc).strftime(HEADER_DATE_FMT),
            "User-Agent": self.config.user_agent,
        }

    def trigger(self, workflow: Workflow) -> Dict:
        workflow_body, body_size = workflow.get_final_json(self.config, is_part_of_bulk=False)
        return self.send(workflow_body)

    def send(self, workflow_body: Dict) -> Dict:
        try:
            headers = self.__get_headers()
            # Signature and Authorization-header
            content_txt, sig = get_request_signature(self.url, 'POST', workflow_body,
                                                     headers, self.config.workspace_secret)
            headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
            # -----
            resp = requests.post(self.url,
                                 data=content_txt.encode('utf-8'),
                                 headers=headers)
        except Exception as ex:
            error_str = ex.__str__()
            return {
                "success": False,
                "status": "fail",
                "status_code": 500,
                "message": error_str,
            }
        else:
            ok_response = resp.status_code // 100 == 2
            if ok_response:
                return {
                    "success": True,
                    "status": "success",
                    "status_code": resp.status_code,
                    "message": resp.text,
                }
            else:
                return {
                    "success": False,
                    "status": "fail",
                    "status_code": resp.status_code,
                    "message": resp.text,
                }



from typing import Dict
import copy
import json
import jsonschema
import traceback

from .constants import (
    WORKFLOW_RUNTIME_KEYS_POTENTIAL_SIZE_IN_BYTES,
    ATTACHMENT_URL_POTENTIAL_SIZE_IN_BYTES,
    ATTACHMENT_UPLOAD_ENABLED, ALLOW_ATTACHMENTS_IN_BULK_API,
)
from .exception import SuprsendValidationError, InputValueError
from .request_schema import _get_schema_validator


def get_apparent_workflow_body_size(body: Dict, is_part_of_bulk: bool) -> int:
    # ---
    extra_bytes = WORKFLOW_RUNTIME_KEYS_POTENTIAL_SIZE_IN_BYTES
    apparent_body = body
    # ---
    if body.get("data") and body["data"].get("$attachments"):
        num_attachments = len(body["data"]["$attachments"])
        if is_part_of_bulk:
            if ALLOW_ATTACHMENTS_IN_BULK_API:
                # if attachment is allowed in bulk api, then calculate size based on whether auto Upload is enabled
                if ATTACHMENT_UPLOAD_ENABLED:
                    # If auto upload enabled, To calculate size, replace attachment size with equivalent url size
                    extra_bytes += num_attachments * ATTACHMENT_URL_POTENTIAL_SIZE_IN_BYTES
                    # -- remove attachments->data key to calculate data size
                    apparent_body = copy.deepcopy(body)
                    for attach_data in apparent_body["data"]["$attachments"]:
                        del attach_data["data"]
                    # ----
                else:
                    # if auto upload is not enabled, attachment data will be passed as it is.
                    pass
            else:
                # If attachment not allowed, then remove data->$attachments before calculating size
                apparent_body = copy.deepcopy(body)
                apparent_body["data"].pop("$attachments", None)
        else:
            if ATTACHMENT_UPLOAD_ENABLED:
                # if auto upload enabled, to calculate size, replace attachment size with equivalent url size
                extra_bytes += num_attachments * ATTACHMENT_URL_POTENTIAL_SIZE_IN_BYTES
                # -- remove attachments->data key to calculate data size
                apparent_body = copy.deepcopy(body)
                for attach_data in apparent_body["data"]["$attachments"]:
                    del attach_data["data"]
                # ----
            else:
                # if auto upload is not enabled, attachment data will be passed as it is.
                pass

    # ---
    body_size = len(json.dumps(apparent_body, ensure_ascii=False).encode('utf-8'))
    apparent_body_size = body_size + extra_bytes
    # --
    return apparent_body_size


def get_apparent_event_size(event: Dict, is_part_of_bulk: bool) -> int:
    # ---
    extra_bytes = 0
    apparent_body = event
    # ---
    if event.get("properties") and event["properties"].get("$attachments"):
        num_attachments = len(event["properties"]["$attachments"])
        if is_part_of_bulk:
            if ALLOW_ATTACHMENTS_IN_BULK_API:
                # if attachment is allowed in bulk api, then calculate size based on whether auto Upload is enabled
                if ATTACHMENT_UPLOAD_ENABLED:
                    # If auto upload enabled, To calculate size, replace attachment size with equivalent url size
                    extra_bytes += num_attachments * ATTACHMENT_URL_POTENTIAL_SIZE_IN_BYTES
                    # -- remove attachments->data key to calculate data size
                    apparent_body = copy.deepcopy(event)
                    for attach_data in apparent_body["properties"]["$attachments"]:
                        del attach_data["data"]
                    # ----
                else:
                    # if auto upload is not enabled, attachment data will be passed as it is.
                    pass
            else:
                # If attachment not allowed, then remove data->$attachments before calculating size
                apparent_body = copy.deepcopy(event)
                apparent_body["properties"].pop("$attachments", None)
        else:
            if ATTACHMENT_UPLOAD_ENABLED:
                # if auto upload enabled, to calculate size, replace attachment size with equivalent url size
                extra_bytes += num_attachments * ATTACHMENT_URL_POTENTIAL_SIZE_IN_BYTES
                # -- remove attachments->data key to calculate data size
                apparent_body = copy.deepcopy(event)
                for attach_data in apparent_body["properties"]["$attachments"]:
                    del attach_data["data"]
                # ----
            else:
                # if auto upload is not enabled, attachment data will be passed as it is.
                pass

    # ---
    body_size = len(json.dumps(apparent_body, ensure_ascii=False).encode('utf-8'))
    apparent_size = body_size + extra_bytes
    # --
    return apparent_size


def get_apparent_identity_event_size(event: Dict) -> int:
    body_size = len(json.dumps(event, ensure_ascii=False).encode('utf-8'))
    return body_size


def get_apparent_list_broadcast_body_size(body: Dict) -> int:
    body_size = len(json.dumps(body, ensure_ascii=False).encode('utf-8'))
    return body_size


def validate_workflow_body_schema(body: Dict) -> Dict:
    # --- In case data is not provided, set it to empty dict
    if body.get("data") is None:
        body["data"] = {}
    if not isinstance(body["data"], dict):
        raise InputValueError("data must be a dictionary")
    # --------------------------------
    schema_validator = _get_schema_validator('workflow')
    try:
        schema_validator.validate(body)
    except jsonschema.exceptions.ValidationError as ve:
        raise SuprsendValidationError(ve.message)
    return body


def validate_workflow_trigger_body_schema(body: Dict) -> Dict:
    # --- In case data is not provided, set it to empty dict
    if body.get("data") is None:
        body["data"] = {}
    if not isinstance(body["data"], dict):
        raise InputValueError("data must be a dictionary")
    # --------------------------------
    schema_validator = _get_schema_validator("workflow_trigger")
    try:
        schema_validator.validate(body)
    except jsonschema.exceptions.ValidationError as ve:
        raise SuprsendValidationError(ve.message)
    return body


def validate_track_event_schema(body: Dict) -> Dict:
    # --- In case props is not provided, set it to empty dict
    if body.get("properties") is None:
        body["properties"] = {}
    # --------------------------------
    schema_validator = _get_schema_validator('event')
    try:
        schema_validator.validate(body)
    except jsonschema.exceptions.ValidationError as ve:
        raise SuprsendValidationError(ve.message)
    return body


def validate_list_broadcast_body_schema(body: Dict) -> Dict:
    # --- In case data is not provided, set it to empty dict
    if body.get("data") is None:
        body["data"] = {}
    if not isinstance(body["data"], dict):
        raise InputValueError("data must be a dictionary")
    # --------------------------------
    schema_validator = _get_schema_validator('list_broadcast')
    try:
        schema_validator.validate(body)
    except jsonschema.exceptions.ValidationError as ve:
        raise SuprsendValidationError(ve.message)
    return body


def invalid_record_json(failed_record, err):
    if isinstance(err, (InputValueError,)):
        err_str = str(err)
    else:
        # includes SuprsendValidationError,
        # OR any other error
        err_str = traceback.format_exc()
    # ------
    rec = {"record": failed_record, "error": err_str, "code": 500}
    return rec


def safe_get(lst, index, default=None):
    """
    method to safely get element from list
    """
    try:
        return lst[index]
    except IndexError:
        return default




class BulkResponse:
    def __init__(self):
        self.status = None
        self.failed_records = []
        self.total = 0
        self.success = 0
        self.failure = 0
        self.warnings = []

    def __str__(self):
        return f"BulkResponse<status: {self.status} | total: {self.total} | success: {self.success} | " \
               f"failure: {self.failure} | warnings: {len(self.warnings)}>"

    def merge_chunk_response(self, ch_resp):
        if not ch_resp:
            return
        # possible status: success/partial/fail
        if self.status is None:
            self.status = ch_resp["status"]
        else:
            if ch_resp["status"] == "partial":
                self.status = "partial"
            elif self.status == "success":
                if ch_resp["status"] == "fail":
                    self.status = "partial"
            elif self.status == "fail":
                if ch_resp["status"] == "success":
                    self.status = "partial"
        self.total += ch_resp.get("total", 0)
        self.success += ch_resp.get("success", 0)
        self.failure += ch_resp.get("failure", 0)
        failed_recs = ch_resp.get("failed_records", [])
        self.failed_records.extend(failed_recs)

    @classmethod
    def empty_chunk_success_response(cls):
        return {
            "status": "success",
            "status_code": 200,
            "total": 0,
            "success": 0,
            "failure": 0,
            "failed_records": [],
            "raw_response": None,
        }

    @classmethod
    def invalid_records_chunk_response(cls, invalid_records):
        return {
            "status": "fail",
            "status_code": 500,
            "total": len(invalid_records),
            "success": 0,
            "failure": len(invalid_records),
            "failed_records": invalid_records,
            "raw_response": None,
        }

    @classmethod
    def parse_bulk_api_v2_response(cls, resp_json: dict):
        total_count = len(resp_json["records"])
        success_count = sum([1 for _ in resp_json["records"] if _["status"] == "success"])
        failure_count = total_count - success_count
        return {
            "status": ("partial" if success_count > 0 else "fail") if failure_count > 0 else "success",
            "total": total_count,
            "success": success_count,
            "failure": failure_count,
            "records": resp_json["records"],
        }



import os
import json
import jsonschema
from .exception import SuprsendMissingSchema, SuprsendInvalidSchema


# Cached json schema
__JSON_SCHEMAS = dict()


def _get_schema_validator(schema_name: str):
    schema_body = __JSON_SCHEMAS.get(schema_name)
    if not schema_body:
        schema_body = __load_json_schema(schema_name)
        __JSON_SCHEMAS[schema_name] = schema_body
    return schema_body


def __load_json_schema(schema_name: str):
    here = os.path.dirname(os.path.abspath(__file__))
    rel_path = "request_json/{}.json".format(schema_name)
    file_path = os.path.join(here, rel_path)
    with open(file_path) as f:
        schema_body = json.load(f)
        if not schema_body:
            raise SuprsendMissingSchema(schema_name)
        try:
            jsonschema.Draft7Validator.check_schema(schema_body)
        except jsonschema.exceptions.SchemaError as se:
            raise SuprsendInvalidSchema(se.message)
        return jsonschema.Draft7Validator(schema_body)



from datetime import datetime, timezone
import requests
from typing import Dict

from .constants import HEADER_DATE_FMT
from .signature import get_request_signature
from .workflow_request import WorkflowTriggerRequest
from .workflow_trigger_bulk import BulkWorkflowTrigger


class WorkflowsApi:
    def __init__(self, config):
        self.config = config
        self.metadata = {"User-Agent": self.config.user_agent}

    def __get_headers(self):
        return {
            "Content-Type": "application/json; charset=utf-8",
            "Date": datetime.now(timezone.utc).strftime(HEADER_DATE_FMT),
            "User-Agent": self.config.user_agent,
        }

    def trigger(self, workflow: WorkflowTriggerRequest) -> Dict:
        workflow_body, body_size = workflow.get_final_json(self.config, is_part_of_bulk=False)
        try:
            headers = self.__get_headers()
            url = "{}trigger/".format(self.config.base_url)
            # Signature and Authorization-header
            content_txt, sig = get_request_signature(url, 'POST', workflow_body,
                                                     headers, self.config.workspace_secret)
            headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
            # -----
            resp = requests.post(url, data=content_txt.encode('utf-8'), headers=headers)
        except Exception as ex:
            error_str = ex.__str__()
            return {
                "success": False,
                "status": "fail",
                "status_code": 500,
                "message": error_str,
                "raw_response": None,
            }
        else:
            ok_response = resp.status_code // 100 == 2
            try:
                resp_json = resp.json()
            except ValueError:
                resp_json = None
            if ok_response:
                return {
                    "success": True,
                    "status": "success",
                    "status_code": resp.status_code,
                    "message": resp_json.get("message_id") if resp_json else resp.text,
                    "raw_response": resp_json,
                }
            else:
                return {
                    "success": False,
                    "status": "fail",
                    "status_code": resp.status_code,
                    "message": resp_json.get("error", {}).get("message") if resp_json else resp.text,
                    "raw_response": resp_json,
                }

    def bulk_trigger_instance(self):
        """
        USAGE:
        supr_client = Suprsend("__workspace_key__", "__workspace_secret__")
        bulk_ins = supr_client.workflows.bulk_trigger_instance()

        # append one by one
        for i in range(0, 10):
            w = WorkflowTriggerRequest(body) # Workflow instance
            bulk_ins.append(w)

        # append many in one call
        all_workflows = [W1, W2, ...] # multiple workflows
        bulk_ins.append(*all_workflows)

        # call trigger
        response = bulk_ins.trigger()

        :return:
        """
        return BulkWorkflowTrigger(self.config)



# Default urls
DEFAULT_URL = "https://hub.suprsend.com/"

# a API call should not have apparent body size of more than 800KB
BODY_MAX_APPARENT_SIZE_IN_BYTES = 800 * 1024  # 800 * 1024
BODY_MAX_APPARENT_SIZE_IN_BYTES_READABLE = "800KB"

# in general url-size wont exceed 2048 chars or 2048 utf-8 bytes
ATTACHMENT_URL_POTENTIAL_SIZE_IN_BYTES = 2100

# few keys added in-flight, amounting to almost 200 bytes increase per workflow-body
WORKFLOW_RUNTIME_KEYS_POTENTIAL_SIZE_IN_BYTES = 200

# max workflow-records in one bulk api call.
MAX_WORKFLOWS_IN_BULK_API = 100
# max event-records in one bulk api call
MAX_EVENTS_IN_BULK_API = 100

ALLOW_ATTACHMENTS_IN_BULK_API = True
ATTACHMENT_UPLOAD_ENABLED = False

# -- single Identity event limit
IDENTITY_SINGLE_EVENT_MAX_APPARENT_SIZE_IN_BYTES = 10 * 1024
IDENTITY_SINGLE_EVENT_MAX_APPARENT_SIZE_IN_BYTES_READABLE = "10KB"

MAX_IDENTITY_EVENTS_IN_BULK_API = 400

# In TZ Format: "%a, %d %b %Y %H:%M:%S %Z"
HEADER_DATE_FMT = "%a, %d %b %Y %H:%M:%S GMT"



from .constants import (
    BODY_MAX_APPARENT_SIZE_IN_BYTES, BODY_MAX_APPARENT_SIZE_IN_BYTES_READABLE,
)
from .exception import InputValueError
from .attachment import get_attachment_json
from .utils import (get_apparent_workflow_body_size, validate_workflow_trigger_body_schema)


class WorkflowTriggerRequest:
    def __init__(self, body, idempotency_key: str = None, tenant_id: str = None, cancellation_key: str = None):
        if not isinstance(body, (dict,)):
            raise InputValueError("WorkflowTriggerRequest body must be a json/dictionary")
        self.body = body
        self.idempotency_key = idempotency_key
        self.tenant_id = tenant_id
        self.cancellation_key = cancellation_key

    def add_attachment(self, file_path: str, file_name: str = None, ignore_if_error: bool = False):
        if self.body.get("data") is None:
            self.body["data"] = {}
        # if body["data"] is not a dict, not raising error while adding attachment.
        if not isinstance(self.body["data"], (dict,)):
            print("WARNING: attachment cannot be added. please make sure body['data'] is a dictionary. "
                  "WorkflowTriggerRequest" + str(self.as_json()))
            return
        # ---
        attachment = get_attachment_json(file_path, file_name, ignore_if_error)
        if not attachment:
            return
        # --- add the attachment to body->data->$attachments
        if self.body["data"].get("$attachments") is None:
            self.body["data"]["$attachments"] = []
        # -----
        self.body["data"]["$attachments"].append(attachment)

    def get_final_json(self, config, is_part_of_bulk: bool = False):
        # add idempotency key in body if present
        if self.idempotency_key:
            self.body["$idempotency_key"] = self.idempotency_key
        if self.tenant_id:
            self.body["tenant_id"] = self.tenant_id
        if self.cancellation_key:
            self.body["cancellation_key"] = self.cancellation_key
        # --
        self.body = validate_workflow_trigger_body_schema(self.body)
        # ---- Check body size
        apparent_size = get_apparent_workflow_body_size(self.body, is_part_of_bulk)
        if apparent_size > BODY_MAX_APPARENT_SIZE_IN_BYTES:
            raise InputValueError(f"workflow body too big - {apparent_size} Bytes, "
                                  f"must not cross {BODY_MAX_APPARENT_SIZE_IN_BYTES_READABLE}")
        # ----
        return self.body, apparent_size

    def as_json(self):
        body_dict = {**self.body}
        if self.idempotency_key:
            body_dict["$idempotency_key"] = self.idempotency_key
        if self.tenant_id:
            body_dict["tenant_id"] = self.tenant_id
        if self.cancellation_key:
            self.body["cancellation_key"] = self.cancellation_key
        # -----
        return body_dict



import copy
from datetime import datetime, timezone
import requests
from typing import List, Dict

from .constants import (
    BODY_MAX_APPARENT_SIZE_IN_BYTES,
    BODY_MAX_APPARENT_SIZE_IN_BYTES_READABLE,
    MAX_EVENTS_IN_BULK_API,
    ALLOW_ATTACHMENTS_IN_BULK_API,
    HEADER_DATE_FMT,
)
from .exception import InputValueError
from .signature import get_request_signature
from .utils import invalid_record_json, safe_get
from .bulk_response import BulkResponse
from .event import Event


class BulkEventsFactory:

    def __init__(self, config):
        self.config = config

    def new_instance(self):
        """
        USAGE:
        supr_client = Suprsend("__workspace_key__", "__workspace_secret__")
        bulk_ins = supr_client.bulk_events.new_instance()

        # append one by one
        for i in range(0, 10):
            event = Event('distinct_id', 'event_name', {}) # create event instance
            bulk_ins.append(event)

        # append many in one call
        all_events = [Event(..), Event(..), ...] # multiple events
        bulk_ins.append(*all_events)

        # call trigger
        response = bulk_ins.trigger()

        :return:
        """
        return BulkEvents(self.config)


class _BulkEventsChunk:
    _chunk_apparent_size_in_bytes = BODY_MAX_APPARENT_SIZE_IN_BYTES
    _chunk_apparent_size_in_bytes_readable = BODY_MAX_APPARENT_SIZE_IN_BYTES_READABLE
    _max_records_in_chunk = MAX_EVENTS_IN_BULK_API

    def __init__(self, config):
        self.config = config
        self.__chunk = []
        self.__url = self.__get_url()
        self.__headers = self.__common_headers()
        #
        self.__running_size = 0
        self.__running_length = 0
        self.response = None

    def __get_url(self):
        url_formatted = "{}v2/bulk/event/".format(self.config.base_url)
        return url_formatted

    def __common_headers(self):
        return {
            "Content-Type": "application/json; charset=utf-8",
            "User-Agent": self.config.user_agent,
        }

    def __dynamic_headers(self):
        return {
            "Date": datetime.now(timezone.utc).strftime(HEADER_DATE_FMT),
        }

    def __add_event_to_chunk(self, event, event_size):
        # First add size, then event to reduce effects of race condition
        self.__running_size += event_size
        self.__chunk.append(event)
        self.__running_length += 1

    def __check_limit_reached(self):
        if self.__running_length >= self._max_records_in_chunk or \
                self.__running_size >= self._chunk_apparent_size_in_bytes:
            return True
        else:
            return False

    def try_to_add_into_chunk(self, event: Dict, event_size: int) -> bool:
        """
        returns whether passed event was able to get added to this chunk or not,
        if true, event gets added to chunk
        :param event:
        :param event_size:
        :return:
        :raises: InputValueError
        """
        if not event:
            return True
        if self.__check_limit_reached():
            return False
        # ---
        if event_size > BODY_MAX_APPARENT_SIZE_IN_BYTES:
            raise InputValueError(f"Event properties too big - {event_size} Bytes, "
                                  f"must not cross {BODY_MAX_APPARENT_SIZE_IN_BYTES_READABLE}")
        # if apparent_size of event crosses limit
        if self.__running_size + event_size > self._chunk_apparent_size_in_bytes:
            return False

        if not ALLOW_ATTACHMENTS_IN_BULK_API:
            event["properties"].pop("$attachments", None)

        # Add Event to chunk
        self.__add_event_to_chunk(event, event_size)
        return True

    def trigger(self):
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(self.__url, 'POST', self.__chunk, headers,
                                                 self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        try:
            resp = requests.post(self.__url,
                                 data=content_txt.encode('utf-8'),
                                 headers=headers)
        except Exception as ex:
            error_str = ex.__str__()
            self.response = {
                "status": "fail",
                "status_code": 500,
                "total": len(self.__chunk),
                "success": 0,
                "failure": len(self.__chunk),
                "failed_records": [{"record": c, "error": error_str, "code": 500} for c in self.__chunk],
                "raw_response": None,
            }
        else:
            # TODO: handle 500/503 errors
            ok_response = resp.status_code // 100 == 2
            resp_json = resp.json()
            parsed_resp = BulkResponse.parse_bulk_api_v2_response(resp_json)
            if ok_response:
                self.response = {
                    "status": parsed_resp["status"],
                    "status_code": resp.status_code,
                    "total": parsed_resp["total"],
                    "success": parsed_resp["success"],
                    "failure": parsed_resp["failure"],
                    "failed_records": [
                        {"record": safe_get(self.__chunk, idx), "error": record["error"]["message"], "code": record["status_code"]}
                        for idx, record in enumerate(resp_json["records"]) if record["status"] == "error"],
                    "raw_response": resp_json
                }
            else:
                self.response = {
                    "status": "fail",
                    "status_code": resp.status_code,
                    "total": len(self.__chunk),
                    "success": 0,
                    "failure": len(self.__chunk),
                    "failed_records": [{"record": c, "error": resp_json["error"]["message"], "code": resp.status_code}
                                       for c in self.__chunk],
                    "raw_response": resp_json
                }


class BulkEvents:
    def __init__(self, config):
        self.config = config
        self.__events = []
        self.__pending_records = []
        self.chunks = []
        self.response = BulkResponse()
        # invalid_record json: {"record": event-json, "error": error_str, "code": 500}
        self.__invalid_records = []

    def __validate_events(self):
        for ev in self.__events:
            try:
                ev_json, body_size = ev.get_final_json(self.config, is_part_of_bulk=True)
                self.__pending_records.append((ev_json, body_size))
            except Exception as ex:
                inv_rec = invalid_record_json(ev.as_json(), ex)
                self.__invalid_records.append(inv_rec)

    def __chunkify(self, start_idx=0):
        curr_chunk = _BulkEventsChunk(self.config)
        self.chunks.append(curr_chunk)
        for rel_idx, rec in enumerate(self.__pending_records[start_idx:]):
            is_added = curr_chunk.try_to_add_into_chunk(rec[0], rec[1])
            if not is_added:
                # create chunks from remaining records
                self.__chunkify(start_idx=(start_idx + rel_idx))
                # Don't forget to break. As current loop must not continue further
                break

    def append(self, *events):
        if not events:
            return
        for ev in events:
            if ev and isinstance(ev, Event):
                ev_copy = copy.deepcopy(ev)
                self.__events.append(ev_copy)

    def trigger(self):
        self.__validate_events()
        # --------
        if len(self.__invalid_records) > 0:
            ch_response = BulkResponse.invalid_records_chunk_response(self.__invalid_records)
            self.response.merge_chunk_response(ch_response)
        # --------
        if len(self.__pending_records):
            self.__chunkify()
            for c_idx, ch in enumerate(self.chunks):
                if self.config.req_log_level > 0:
                    print(f"DEBUG: triggering api call for chunk: {c_idx}")
                # do api call
                ch.trigger()
                # merge response
                self.response.merge_chunk_response(ch.response)
        else:
            # if no records. i.e. len(invalid_records) and len(pending_records) both are 0
            # then add empty success response
            if len(self.__invalid_records) == 0:
                self.response.merge_chunk_response(BulkResponse.empty_chunk_success_response())
        # -----
        return self.response



from datetime import datetime, timezone
import requests
import time
from typing import List, Dict
import uuid

from .constants import (
    HEADER_DATE_FMT,
    BODY_MAX_APPARENT_SIZE_IN_BYTES, BODY_MAX_APPARENT_SIZE_IN_BYTES_READABLE,
)
from .exception import InputValueError
from .attachment import get_attachment_json
from .signature import get_request_signature
from .utils import (validate_track_event_schema, get_apparent_event_size, )


RESERVED_EVENT_NAMES = [
    "$identify",
    "$notification_delivered", "$notification_dismiss", "$notification_clicked",
    "$app_launched", "$user_login", "$user_logout",
]


class Event:
    def __init__(self, distinct_id: str, event_name: str, properties: Dict = None,
                 idempotency_key: str = None, tenant_id: str = None, brand_id: str = None):
        self.distinct_id = distinct_id
        self.event_name = event_name
        self.properties = properties
        self.idempotency_key = idempotency_key
        self.tenant_id = tenant_id
        self.brand_id = brand_id
        # default values
        if self.properties is None:
            self.properties = {}

    def __validate_distinct_id(self):
        if not isinstance(self.distinct_id, (str,)):
            raise InputValueError("distinct_id must be a string. an Id which uniquely identify a user in your app")
        distinct_id = self.distinct_id.strip()
        if not distinct_id:
            raise InputValueError("distinct_id missing")
        self.distinct_id = distinct_id

    def __validate_properties(self):
        if not isinstance(self.properties, (dict,)):
            raise InputValueError("properties must be a dictionary")

    def __check_event_prefix(self, event_name: str):
        if event_name not in RESERVED_EVENT_NAMES:
            prefix_3_chars = event_name[:3].lower()
            if prefix_3_chars.startswith("$") or prefix_3_chars == "ss_":
                raise InputValueError("event_names starting with [$,ss_] are reserved by SuprSend")

    def __validate_event_name(self):
        if not isinstance(self.event_name, (str,)):
            raise InputValueError("event_name must be a string")
        event_name = self.event_name.strip()
        if not event_name:
            raise InputValueError("event_name missing")
        self.__check_event_prefix(event_name)
        self.event_name = event_name

    def add_attachment(self, file_path: str, file_name: str = None, ignore_if_error: bool = False):
        # if properties is not a dict, not raising error while adding attachment.
        if not isinstance(self.properties, (dict,)):
            print("WARNING: attachment cannot be added. please make sure properties is a dictionary. Event" +
                  str(self.as_json()))
            return
        # ---
        attachment = get_attachment_json(file_path, file_name, ignore_if_error)
        if not attachment:
            return
        # --- add the attachment to properties->$attachments
        if self.properties.get("$attachments") is None:
            self.properties["$attachments"] = []
        # -----
        self.properties["$attachments"].append(attachment)

    def get_final_json(self, config, is_part_of_bulk: bool = False):
        # --- validate
        self.__validate_distinct_id()
        self.__validate_event_name()
        self.__validate_properties()
        # ---
        super_props = {"$ss_sdk_version": config.user_agent}
        event_dict = {
            "$insert_id": str(uuid.uuid4()),
            "$time": int(time.time() * 1000),
            "event": self.event_name,
            "env": config.workspace_key,
            "distinct_id": self.distinct_id,
            "properties": {**self.properties, **super_props}
        }
        if self.idempotency_key:
            event_dict["$idempotency_key"] = self.idempotency_key
        if self.tenant_id:
            event_dict["tenant_id"] = self.tenant_id
        if self.brand_id:
            event_dict["brand_id"] = self.brand_id
        # ---
        event_dict = validate_track_event_schema(event_dict)
        # ---- Check size
        apparent_size = get_apparent_event_size(event_dict, is_part_of_bulk)
        if apparent_size > BODY_MAX_APPARENT_SIZE_IN_BYTES:
            raise InputValueError(f"Event size too big - {apparent_size} Bytes, "
                                  f"must not cross {BODY_MAX_APPARENT_SIZE_IN_BYTES_READABLE}")
        # ----
        return event_dict, apparent_size

    def as_json(self):
        event_dict = {
            "event": self.event_name,
            "distinct_id": self.distinct_id,
            "properties": self.properties,
        }
        if self.idempotency_key:
            event_dict["$idempotency_key"] = self.idempotency_key
        if self.tenant_id:
            event_dict["tenant_id"] = self.tenant_id
        if self.brand_id:
            event_dict["brand_id"] = self.brand_id
        # -----
        return event_dict


class EventCollector:
    def __init__(self, config):
        self.config = config
        self.__url = self.__get_url()
        self.__headers = self.__common_headers()

    def __get_url(self):
        url_formatted = "{}v2/event/".format(self.config.base_url)
        return url_formatted

    def __common_headers(self):
        return {
            "Content-Type": "application/json; charset=utf-8",
            "User-Agent": self.config.user_agent,
        }

    def __dynamic_headers(self):
        return {
            "Date": datetime.now(timezone.utc).strftime(HEADER_DATE_FMT),
        }

    def collect(self, event: Event) -> Dict:
        event_dict, event_size = event.get_final_json(self.config, is_part_of_bulk=False)
        return self.send(event_dict)

    def send(self, event: Dict) -> Dict:
        try:
            headers = {**self.__headers, **self.__dynamic_headers()}
            # Signature and Authorization-header
            content_txt, sig = get_request_signature(self.__url, 'POST', event, headers,
                                                     self.config.workspace_secret)
            headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
            # -----
            resp = requests.post(self.__url,
                                 data=content_txt.encode('utf-8'),
                                 headers=headers)
        except Exception as ex:
            error_str = ex.__str__()
            return {
                "success": False,
                "status": "fail",
                "status_code": 500,
                "message": error_str,
                "raw_response": None,
            }
        else:
            ok_response = resp.status_code // 100 == 2
            resp_json = resp.json()
            if ok_response:
                return {
                    "success": True,
                    "status": "success",
                    "status_code": resp.status_code,
                    "message": resp_json.get("message_id"),
                    "raw_response": resp_json,
                }
            else:
                return {
                    "success": False,
                    "status": "fail",
                    "status_code": resp.status_code,
                    "message": resp_json.get("error", {}).get("message"),
                    "raw_response": resp_json,
                }



from datetime import datetime, timezone
from typing import Dict, Union
import requests
import urllib.parse

from .constants import (
    HEADER_DATE_FMT,
)
from .exception import SuprsendAPIException, SuprsendValidationError
from .signature import get_request_signature
from .user_edit import UserEdit
from .users_edit_bulk import BulkUsersEdit


class UsersApi:
    def __init__(self, config):
        self.config = config
        self.list_url = "{}v1/user/".format(self.config.base_url)
        self.bulk_url = "{}v1/bulk/user/".format(self.config.base_url)

    def __get_headers(self):
        return {
            "Content-Type": "application/json; charset=utf-8",
            "User-Agent": self.config.user_agent,
            "Date": datetime.now(timezone.utc).strftime(HEADER_DATE_FMT),
        }

    def list(self, options: Dict = None) -> Dict:
        encoded_options = urllib.parse.urlencode((options or {}))
        url = "{}{}".format(self.list_url, (f"?{encoded_options}" if encoded_options else ""))
        headers = self.__get_headers()
        # ---
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "GET", None, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.get(url, headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def _validate_distinct_id(self, distinct_id: str) -> str:
        if not distinct_id or not isinstance(distinct_id, (str,)) or not distinct_id.strip():
            raise SuprsendValidationError("missing distinct_id")
        return distinct_id.strip()

    def detail_url(self, distinct_id: str) -> str:
        distinct_id = self._validate_distinct_id(distinct_id)
        distinct_id_encoded = urllib.parse.quote_plus(distinct_id)
        return f"{self.list_url}{distinct_id_encoded}/"

    def get(self, distinct_id: str) -> Dict:
        url = self.detail_url(distinct_id)
        headers = self.__get_headers()
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "GET", None, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.get(url, headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def upsert(self, distinct_id: str, payload: Dict = None) -> Dict:
        url = self.detail_url(distinct_id)
        payload = payload or {}
        headers = self.__get_headers()
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "POST", payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.post(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def async_edit(self, edit_instance: UserEdit) -> Dict:
        if not edit_instance:
            raise SuprsendValidationError("instance is required")
        edit_instance.validate_body()
        a_payload = edit_instance.get_async_payload()
        edit_instance.validate_payload_size(a_payload)
        # --- Signature and Authorization-header
        url = "{}event/".format(self.config.base_url)
        headers = self.__get_headers()
        content_txt, sig = get_request_signature(url, "POST", a_payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.post(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        # if no error, return success response
        return {"success": True, "status": "success", "status_code": resp.status_code, "message": resp.text}

    def edit(self, edit_ins_or_distinct_id: Union[UserEdit, str], edit_payload: Dict = None) -> Dict:
        if isinstance(edit_ins_or_distinct_id, UserEdit):
            edit_ins = edit_ins_or_distinct_id
            edit_ins.validate_body()
            payload = edit_ins.get_payload()
            url = self.detail_url(edit_ins.distinct_id)
        else:
            distinct_id = edit_ins_or_distinct_id
            payload = edit_payload or {}
            url = self.detail_url(distinct_id)
        # ----
        headers = self.__get_headers()
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "PATCH", payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.patch(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def merge(self, distinct_id: str, from_user_id: str) -> Dict:
        url = "{}merge/".format(self.detail_url(distinct_id))
        payload = {"from_user_id": from_user_id}
        headers = self.__get_headers()
        # ---
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "POST", payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.post(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def delete(self, distinct_id: str) -> Dict:
        url = self.detail_url(distinct_id)
        headers = self.__get_headers()
        # ---
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "DELETE", "", headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.delete(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return {"success": True, "status_code": resp.status_code}

    def bulk_delete(self, payload: Dict) -> Dict:
        """
        payload: {"distinct_ids": ["id1", "id2"]}
        :param payload:
        :return:
        """
        payload = payload or {}
        url = self.bulk_url
        headers = self.__get_headers()
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "DELETE", payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.delete(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return {"success": True, "status_code": resp.status_code}

    def get_objects_subscribed_to(self, distinct_id: str, options: Dict = None) -> Dict:
        encoded_options = urllib.parse.urlencode((options or {}))
        _detail_url = self.detail_url(distinct_id)
        url = "{}subscribed_to/object/{}".format(_detail_url, (f"?{encoded_options}" if encoded_options else ""))
        headers = self.__get_headers()
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "GET", None, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.get(url, headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def get_lists_subscribed_to(self, distinct_id: str, options: Dict = None) -> Dict:
        encoded_options = urllib.parse.urlencode((options or {}))
        _detail_url = self.detail_url(distinct_id)
        url = "{}subscribed_to/list/{}".format(_detail_url, (f"?{encoded_options}" if encoded_options else ""))
        headers = self.__get_headers()
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, "GET", None, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.get(url, headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def get_edit_instance(self, distinct_id: str) -> UserEdit:
        distinct_id = self._validate_distinct_id(distinct_id)
        return UserEdit(self.config, distinct_id)

    def get_bulk_edit_instance(self) -> BulkUsersEdit:
        return BulkUsersEdit(self.config)



from datetime import datetime, timezone
import requests
import time
from typing import List, Dict
import urllib.parse
import uuid

from .exception import InputValueError, SuprsendAPIException, SuprsendValidationError
from .constants import (
    HEADER_DATE_FMT,
    BODY_MAX_APPARENT_SIZE_IN_BYTES, BODY_MAX_APPARENT_SIZE_IN_BYTES_READABLE,
)
from .utils import (get_apparent_list_broadcast_body_size, validate_list_broadcast_body_schema)
from .signature import get_request_signature
from .attachment import get_attachment_json


class SubscriberListBroadcast:
    def __init__(self, body, idempotency_key: str = None, tenant_id: str = None, brand_id: str = None):
        if not isinstance(body, (dict,)):
            raise InputValueError("broadcast body must be a json/dictionary")
        self.body = body
        self.idempotency_key = idempotency_key
        self.tenant_id = tenant_id
        self.brand_id = brand_id

    def add_attachment(self, file_path: str, file_name: str = None, ignore_if_error: bool = False):
        if self.body.get("data") is None:
            self.body["data"] = {}
        # if body["data"] is not a dict, not raising error while adding attachment.
        if not isinstance(self.body["data"], (dict,)):
            print("WARNING: attachment cannot be added. please make sure body['data'] is a dictionary. "
                  "SubscriberListBroadcast" + str(self.as_json()))
            return
        # ---
        attachment = get_attachment_json(file_path, file_name, ignore_if_error)
        if not attachment:
            return
        # --- add the attachment to body->data->$attachments
        if self.body["data"].get("$attachments") is None:
            self.body["data"]["$attachments"] = []
        # -----
        self.body["data"]["$attachments"].append(attachment)

    def get_final_json(self):
        self.body["$insert_id"] = str(uuid.uuid4())
        self.body["$time"] = int(time.time() * 1000)
        if self.idempotency_key:
            self.body["$idempotency_key"] = self.idempotency_key
        if self.tenant_id:
            self.body["tenant_id"] = self.tenant_id
        if self.brand_id:
            self.body["brand_id"] = self.brand_id
        # --
        self.body = validate_list_broadcast_body_schema(self.body)
        # ---- Check body size
        apparent_size = get_apparent_list_broadcast_body_size(self.body)
        if apparent_size > BODY_MAX_APPARENT_SIZE_IN_BYTES:
            raise InputValueError(f"SubscriberListBroadcast body too big - {apparent_size} Bytes, "
                                  f"must not cross {BODY_MAX_APPARENT_SIZE_IN_BYTES_READABLE}")
        # ----
        return self.body, apparent_size

    def as_json(self):
        body_dict = {**self.body}
        if self.idempotency_key:
            body_dict["$idempotency_key"] = self.idempotency_key
        if self.tenant_id:
            body_dict["tenant_id"] = self.tenant_id
        if self.brand_id:
            body_dict["brand_id"] = self.brand_id
        # -----
        return body_dict


class SubscriberListsApi:
    def __init__(self, config):
        self.config = config
        self.subscriber_list_url = "{}v1/subscriber_list/".format(self.config.base_url)
        self.broadcast_url = "{}{}/broadcast/".format(self.config.base_url, self.config.workspace_key)
        self.__headers = self.__common_headers()
        self.non_error_default_response = {"success": True}

    def __common_headers(self):
        return {
            "Content-Type": "application/json; charset=utf-8",
            "User-Agent": self.config.user_agent,
        }

    def __dynamic_headers(self):
        return {
            "Date": datetime.now(timezone.utc).strftime(HEADER_DATE_FMT),
        }

    def _validate_list_id(self, list_id):
        if not isinstance(list_id, (str,)):
            raise SuprsendValidationError("list_id must be a string")
        list_id = list_id.strip()
        if not list_id:
            raise SuprsendValidationError("missing list_id")
        return list_id

    def create(self, payload: Dict):
        if not payload:
            raise SuprsendValidationError("missing payload")
        list_id = payload.get("list_id")
        if not list_id:
            raise SuprsendValidationError("missing list_id is payload")
        list_id = self._validate_list_id(list_id)
        # -----
        payload["list_id"] = list_id
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(self.subscriber_list_url, 'POST', payload, headers,
                                                 self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.post(self.subscriber_list_url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def cleaned_limit_offset(self, limit: int, offset: int):
        # limit must be 0 < x <= 1000
        limit = limit if (isinstance(limit, int) and 0 < limit <= 1000) else 20
        # offset must be >=0
        offset = offset if (isinstance(offset, int) and offset >= 0) else 0
        #
        return limit, offset

    def get_all(self, limit: int = 20, offset: int = 0):
        limit, offset = self.cleaned_limit_offset(limit, offset)
        params = {"limit": limit, "offset": offset}
        encoded_params = urllib.parse.urlencode(params)
        #
        url = f"{self.subscriber_list_url}?{encoded_params}"
        # ---
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'GET', None, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.get(url, headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def __subscriber_list_detail_url(self, list_id: str):
        list_id = str(list_id).strip()
        list_id_encoded = urllib.parse.quote_plus(list_id)
        url = f"{self.subscriber_list_url}{list_id_encoded}/"
        return url

    def get(self, list_id: str):
        list_id = self._validate_list_id(list_id)
        # --------
        url = self.__subscriber_list_detail_url(list_id)
        # ---
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'GET', None, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.get(url, headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def add(self, list_id: str, distinct_ids: list):
        list_id = self._validate_list_id(list_id)
        if not isinstance(distinct_ids, (list, )):
            raise SuprsendValidationError("distinct_ids must be list of strings")
        if len(distinct_ids) == 0:
            return self.non_error_default_response
        url = "{}subscriber/add/".format(self.__subscriber_list_detail_url(list_id))
        payload = {"distinct_ids": distinct_ids}
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'POST', payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.post(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def remove(self, list_id: str, distinct_ids: list):
        list_id = self._validate_list_id(list_id)
        if not isinstance(distinct_ids, (list,)):
            raise SuprsendValidationError("distinct_ids must be list of strings")
        if len(distinct_ids) == 0:
            return self.non_error_default_response
        url = "{}subscriber/remove/".format(self.__subscriber_list_detail_url(list_id))
        payload = {"distinct_ids": distinct_ids}
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'POST', payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.post(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def delete(self, list_id: str):
        list_id = self._validate_list_id(list_id)
        url = "{}delete/".format(self.__subscriber_list_detail_url(list_id))
        headers = {**self.__headers, **self.__dynamic_headers()}
        payload = {}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'PATCH', payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.patch(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def broadcast(self, broadcast_instance: SubscriberListBroadcast) -> Dict:
        if not isinstance(broadcast_instance, SubscriberListBroadcast):
            raise InputValueError("argument must be an instance of suprsend.SubscriberListBroadcast")

        broadcast_body, body_size = broadcast_instance.get_final_json()
        try:
            headers = {**self.__headers, **self.__dynamic_headers()}
            # Signature and Authorization-header
            content_txt, sig = get_request_signature(self.broadcast_url, 'POST', broadcast_body,
                                                     headers, self.config.workspace_secret)
            headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
            # -----
            resp = requests.post(self.broadcast_url,
                                 data=content_txt.encode('utf-8'),
                                 headers=headers)
        except Exception as ex:
            error_str = ex.__str__()
            return {
                "success": False,
                "status": "fail",
                "status_code": 500,
                "message": error_str,
            }
        else:
            ok_response = resp.status_code // 100 == 2
            if ok_response:
                return {
                    "success": True,
                    "status": "success",
                    "status_code": resp.status_code,
                    "message": resp.text,
                }
            else:
                return {
                    "success": False,
                    "status": "fail",
                    "status_code": resp.status_code,
                    "message": resp.text,
                }

    def start_sync(self, list_id: str):
        list_id = self._validate_list_id(list_id)

        url = "{}start_sync/".format(self.__subscriber_list_detail_url(list_id))
        payload = {}
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'POST', payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.post(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def _validate_version_id(self, version_id):
        if not isinstance(version_id, (str,)):
            raise SuprsendValidationError("version_id must be a string")
        version_id = version_id.strip()
        if not version_id:
            raise SuprsendValidationError("missing version_id")
        return version_id

    def __subscriber_list_url_with_version(self, list_id: str, version_id: str):
        list_id = str(list_id).strip()
        list_id_encoded = urllib.parse.quote_plus(list_id)
        version_id = str(version_id).strip()
        version_id_encoded = urllib.parse.quote_plus(version_id)
        url = f"{self.subscriber_list_url}{list_id_encoded}/version/{version_id_encoded}/"
        return url

    def get_version(self, list_id: str, version_id: str):
        list_id = self._validate_list_id(list_id)
        version_id = self._validate_version_id(version_id)
        # --------
        url = self.__subscriber_list_url_with_version(list_id, version_id)
        # ---
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'GET', None, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.get(url, headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def add_to_version(self, list_id: str, version_id: str, distinct_ids: list):
        list_id = self._validate_list_id(list_id)
        if not isinstance(distinct_ids, (list,)):
            raise SuprsendValidationError("distinct_ids must be list of strings")
        if len(distinct_ids) == 0:
            return self.non_error_default_response

        version_id = self._validate_version_id(version_id)
        url = "{}subscriber/add/".format(self.__subscriber_list_url_with_version(list_id, version_id))
        payload = {"distinct_ids": distinct_ids}
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'POST', payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.post(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def remove_from_version(self, list_id: str, version_id: str, distinct_ids: list):
        list_id = self._validate_list_id(list_id)
        if not isinstance(distinct_ids, (list,)):
            raise SuprsendValidationError("distinct_ids must be list of strings")
        if len(distinct_ids) == 0:
            return self.non_error_default_response
        version_id = self._validate_version_id(version_id)
        url = "{}subscriber/remove/".format(self.__subscriber_list_url_with_version(list_id, version_id))
        payload = {"distinct_ids": distinct_ids}
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'POST', payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.post(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def finish_sync(self, list_id: str, version_id: str):
        list_id = self._validate_list_id(list_id)
        version_id = self._validate_version_id(version_id)
        url = "{}finish_sync/".format(self.__subscriber_list_url_with_version(list_id, version_id))
        payload = {}
        headers = {**self.__headers, **self.__dynamic_headers()}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'PATCH', payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.patch(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()

    def delete_version(self, list_id: str, version_id: str):
        list_id = self._validate_list_id(list_id)
        version_id = self._validate_version_id(version_id)

        url = "{}delete/".format(self.__subscriber_list_url_with_version(list_id, version_id))
        headers = {**self.__headers, **self.__dynamic_headers()}
        payload = {}
        # Signature and Authorization-header
        content_txt, sig = get_request_signature(url, 'PATCH', payload, headers, self.config.workspace_secret)
        headers["Authorization"] = "{}:{}".format(self.config.workspace_key, sig)
        # -----
        resp = requests.patch(url, data=content_txt.encode('utf-8'), headers=headers)
        if resp.status_code >= 400:
            raise SuprsendAPIException(resp)
        return resp.json()



import platform

from typing import List, Dict
from warnings import warn

from .version import __version__
from .constants import DEFAULT_URL
from .exception import SuprsendConfigError, InputValueError
from .attachment import get_attachment_json
from .workflow import Workflow, _WorkflowTrigger
from .workflow_api import WorkflowsApi
from .request_log import set_logging
from .workflows_bulk import BulkWorkflowsFactory
from .events_bulk import BulkEventsFactory
from .subscribers_bulk import BulkSubscribersFactory
from .subscriber import SubscriberFactory
from .subscriber_list import SubscriberListsApi
from .event import Event, EventCollector
from .tenant import TenantsApi
from .brand import BrandsApi
from .objects_api import ObjectsApi
from .users_api import UsersApi


class Suprsend:
    """
    - Basic instance
     supr_client = Suprsend("__workspace_key__", "__workspace_secret__")
    - Instance with debug on
     supr_client = Suprsend("__workspace_key__", "__workspace_secret__", debug=True)
    - Instance with custom base-url
     supr_client = Suprsend("__workspace_key__", "__workspace_secret__", base_url="https://example.com/", debug=False)
    """
    def __init__(self, workspace_key: str, workspace_secret: str, base_url: str = None, debug: bool = False, **kwargs):
        self.workspace_key = workspace_key
        self.workspace_secret = workspace_secret
        #
        self.sdk_version = __version__
        self.user_agent = "suprsend/{};python/{}".format(self.sdk_version, platform.python_version())
        #
        self.base_url = self.__get_base_url(base_url)
        # ---
        self.__validate()
        # --- set logging level for http request
        self.req_log_level = 1 if debug else 0
        set_logging(self.req_log_level)
        #
        self._workflow_trigger = _WorkflowTrigger(self)
        self._eventcollector = EventCollector(self)
        # -- bulk instances
        self._bulk_workflows = BulkWorkflowsFactory(self)
        self._bulk_events = BulkEventsFactory(self)
        self._bulk_users = BulkSubscribersFactory(self)
        # --
        self._user = SubscriberFactory(self)
        # --
        self.tenants = TenantsApi(self)
        self.brands = BrandsApi(self)
        self.workflows = WorkflowsApi(self)
        self.objects = ObjectsApi(self)
        self.users = UsersApi(self)
        # --
        self.subscriber_lists = SubscriberListsApi(self)

    @property
    def bulk_workflows(self):
        return self._bulk_workflows

    @property
    def bulk_events(self):
        return self._bulk_events

    @property
    def bulk_users(self):
        return self._bulk_users

    @property
    def user(self):
        return self._user

    @staticmethod
    def __get_base_url(base_url):
        # ---- strip
        if base_url:
            base_url = base_url.strip()
        # ---- if url not passed, set url based on server env
        if not base_url:
            base_url = DEFAULT_URL
        # ---- check url ends with /
        base_url = base_url.strip()
        if base_url[len(base_url) - 1] != "/":
            base_url = base_url + "/"
        return base_url

    def __validate(self):
        if not self.workspace_key:
            raise SuprsendConfigError("Missing workspace_key")
        if not self.workspace_secret:
            raise SuprsendConfigError("Missing workspace_secret")
        if not self.base_url:
            raise SuprsendConfigError("Missing base_url")

    def add_attachment(self, body: Dict, file_path: str, file_name: str = None, ignore_if_error: bool = False) -> Dict:
        warn('This method is deprecated. Use "WorkflowTriggerRequest.add_attachment()" instead',
             DeprecationWarning, stacklevel=2)
        # if data key not present, add it and set value={}.
        if body.get("data") is None:
            body["data"] = {}
        if not isinstance(body, dict):
            raise InputValueError("data must be a dictionary")
        # --------
        attachment = get_attachment_json(file_path, file_name, ignore_if_error)
        # --- add the attachment to body->data->$attachments
        if body["data"].get("$attachments") is None:
            body["data"]["$attachments"] = []
        #
        body["data"]["$attachments"].append(attachment)
        return body

    def trigger_workflow(self, data) -> Dict:
        """
        :param data:
        :return: {
            "success": true/false,
            "status": "success"/"fail",
            "message": "message",
            "status_code": 202/401/500,
        }
        :except:
            - SuprsendValidationError
        """
        # warn('This method will be deprecated. Use client.workflows.trigger(WorkflowTriggerRequest) instead',
        #      DeprecationWarning, stacklevel=2)
        if isinstance(data, Workflow):
            wf_ins = data
        else:
            wf_ins = Workflow(data, idempotency_key=None, tenant_id=None)
        # -----
        return self._workflow_trigger.trigger(wf_ins)

    def track(self, distinct_id: str, event_name: str, properties: Dict = None,
              idempotency_key: str = None, tenant_id: str = None, brand_id: str = None) -> Dict:
        """
        :param distinct_id:
        :param event_name:
        :param properties:
        :param idempotency_key:
        :param tenant_id:
        :param brand_id:
        :return: {
            "success": True,
            "status": "success",
            "status_code": resp.status_code,
            "message": resp.text,
        }
        :except:
            - SuprsendValidationError (if post-data is invalid.)
            - ValueError
        """
        warn('This method is deprecated. Use "track_event(Event)" instead', DeprecationWarning, stacklevel=2)
        if not tenant_id:
            tenant_id = brand_id
        # ---
        event = Event(distinct_id, event_name, properties, idempotency_key=idempotency_key, tenant_id=tenant_id)
        return self._eventcollector.collect(event)

    def track_event(self, event: Event) -> Dict:
        """
        :param event: suprsend.Event
        :return: {
            "success": True,
            "status": "success",
            "status_code": resp.status_code,
            "message": resp.text,
        }
        :except:
            - SuprsendValidationError (if post-data is invalid.)
            - ValueError
        """
        if not isinstance(event, Event):
            raise InputValueError("argument must be an instance of suprsend.Event")
        return self._eventcollector.collect(event)



class SuprsendError(Exception):
    """
    A base class for all Suprsend exceptions.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        self.message = kwargs.get('message')
        if self.message is None and self.args and self.args[0]:
            self.message = self.args[0]
        elif self.message is None:
            self.message = None
        self.status_code = kwargs.get('status_code', 500)

    def __str__(self):
        return f"[code: {self.status_code}] {self.message}"


class SuprsendAPIException(SuprsendError):
    response = None
    message = "unknown error"

    def __init__(self, response):
        # -- Get error message
        content_type = response.headers.get("Content-Type")
        if content_type and "application/json" in content_type:
            try:
                rjson = response.json()
                err_msg = rjson.get("message") or rjson.get("detail")
            except:
                err_msg = response.text
        else:
            err_msg = response.text
        message = f"{err_msg}"
        # --
        super().__init__(message=message, status_code=response.status_code)
        self.response = response


class InputValueError(ValueError):
    pass


class SuprsendConfigError(SuprsendError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SuprsendMissingSchema(SuprsendError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.status_code = 500
        self.error_type = "SuprsendMissingSchema"
        if self.message is None:
            self.message = "Missing JSON schema"


class SuprsendInvalidSchema(SuprsendError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.status_code = 500
        self.error_type = "SuprsendInvalidSchema"
        if self.message is None:
            self.message = "Invalid JSON schema"


class SuprsendValidationError(SuprsendError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.status_code = 400
        self.error_type = "SuprsendValidationError"
        if self.message is None:
            self.message = "validation error"
